let () = Init.init_netdb ()

include Glue
include Pipe

type to_do =
| Depending_install
| Depending_upgrade
| Depending_remove
| Depending_reinstall
| Install
| Upgrade
| Remove
| Reinstall
| Dont_touch

type glue = (<
             get_packages : Glue.package array;
             get_wishes: Glue.wish list ;
             check_wishes:  ?force_rescan: bool -> Glue.wish list -> Glue.plan_t ;
             get_dependencies: options:( string * string ) list -> string -> fdependencies;
             remember_config: options:(string * string) list -> string -> unit ;
             get_apply_order : Glue.wish list -> Glue.plan_t ;
             get_bin_fln:  Glue.binary_identifier -> string option ;
             get_newer: unit -> string list
                        >)

let (//) = Filename.concat

module Const = struct
  let buffer_len = 8192

  let invalid_string = "\000\000"

  let to_install_color = "#88ee22"
  let to_install_dependent_color = "#eeff11"

  let to_remove_color = "#ee3322"
  let to_remove_dependent_color = "#ff8888"

  let categories = [
    "apps","apps (end user software)" ;
    "base","base packages";
    "conf","configuration packages";
    "godi","godi (libraries,...)";
    "oasis","OASIS-DB Imports";
  ]

  let status_cat_updatable =  "available updates"
  let status_cat_installed =  "installed"
  let status_cat_not_installed = "not installed"

  let prog_name = Version.prog_name

  let windows_platform = match Sys.os_type with
  | "Win32" | "Win64" -> true
  | _ -> false

  let godi_dir,sub_prog_env = match windows_platform with
  | false -> Version.godi_dir,None
  | true ->
  (* canonical install location is /opt/[gw]odiXYZ/gui/bin/gui.exe,
     We have to make sure, that PATH contains all neccessary paths if this program is called
     from outside cygwin.
     assuming default godi/cygwin paths:
     /opt/ [gw]odiXYZ/bin:/opt/[gw]odiXYZ/sbin:/usr/local/bin:/usr/bin:/bin:...windows paths
     TODO: The user could intentionally change the order of the above paths
  *)
    let godi_dir =
      let canonic = Filename.dirname ^$ Filename.dirname ^$ Filename.dirname Util.executable_name in
      if Sys.file_exists canonic = true then
        canonic
      else
        try
          let t = Sys.getenv "GODI_PREFIX" in
          if Sys.file_exists t then
            t
          else
            raise Not_found
        with
        | Not_found -> Glib.Message.log `CRITICAL "I could not determine the location of your Godi installation. \
Please move this program to the right location or set GODI_PREFIX";
          exit 1;
    in
    let cygwin_root = Filename.dirname ^$ Filename.dirname godi_dir in
    let godi_bin =  godi_dir // "bin"
    and godi_sbin = godi_dir // "sbin"
    and usr_local_bin = cygwin_root // "usr" // "local" // "bin"
    and usr_bin = cygwin_root // "usr" // "bin"
    and bin = cygwin_root // "bin" in
    let old_path =
      try Sys.getenv "PATH"
      with Not_found ->
        try Sys.getenv "Path" with Not_found -> ""
    in
    let (invalid_paths,cur_paths) = List.partition Util.contains_spaces ^$ Util.split old_path ';' in
    (match invalid_paths with
    | [] -> ()
    | paths ->
      let window_title = "Invalid PATH"
      and title = "Some paths will be ignored"
      and info = "The PATH variable contains spaces.\
This is not supported by this program. The problematic PATH components are removed \
for the run time of this program. Beware that this might cause other problems.\n";
      and tree = ["paths that are ignored ",true,paths] in
      Info_tree.show_error ~window_title ~title ~info ~tree ()
    );
    let cur_paths = Util.list_unique (godi_bin::godi_sbin::usr_local_bin::usr_bin::bin::cur_paths) |>
        String.concat ";"
    in
    if cur_paths = old_path then
      godi_dir,None
    else (
      let new_env =
        Array.append
          (Util.environment_with_removed_values ["PATH";"Path"])
          [| "PATH=" ^ cur_paths   |]
      in
      Unix.putenv "PATH" cur_paths;
      (*Unix.putenv "Path" cur_paths;*)
      godi_dir, Some(new_env)
    )

  let pixbuf_dir =
    let test_file = "package-installed-updated.png"
    and dir1 =  godi_dir // "share" // prog_name // "pixmaps" (* unix location *)
    and dir2 =  "." // "pixmaps" (* development location *)
    and dir3 = godi_dir // "gui" // "share" // "pixmaps" in (*windows binary location *)
    if Sys.file_exists ( dir1 // test_file ) then
      Some dir1
    else if Sys.file_exists ( dir2 // test_file ) then
      Some dir2
    else if Sys.file_exists ( dir3 // test_file ) then
      Some dir3
    else if Sys.file_exists ( "." // test_file ) then
      Some "."
    else
      None

  let help_pixbuf  x = match pixbuf_dir with
  | None -> GdkPixbuf.create ~width:1 ~height:1 ()
  | Some y ->
    try
      GdkPixbuf.from_file ( y // x )
    with
    | _ ->
      (*Glib.Message.log `WARNING "pixbuf %s not found" x ;*)
      GdkPixbuf.create ~width:1 ~height:1 ()

  let pixbuf_package_installed_and_updated =  help_pixbuf  "package-installed-updated.png"
  let pixbuf_package_available = help_pixbuf  "package-available.png"
  let pixbuf_package_install = help_pixbuf  "package-install.png"
  let pixbuf_package_upgrade = help_pixbuf  "package-upgrade.png"
  let pixbuf_package_remove = help_pixbuf  "package-remove.png"
  let pixbuf_package_installed_outdated = help_pixbuf  "package-installed-outdated.png"
  let pixbuf_update = help_pixbuf  "update.png"
  let pixbuf_select_newer = help_pixbuf  "select-newer.png"
  let pixbuf_camel = help_pixbuf "camel48.png"

  let dont_show_source_packages =
    if Util.array_contains Sys.argv "--binary-only" then
      true
    else
      match windows_platform with
      | false -> false
      | true ->
      (* Quick and dirty test, if the user has already called godi_update from console *)
        not ( Sys.file_exists ( godi_dir // "build" // "godi" // "godi-tools" ) )

  let godi_console =
    let exe_suffix = match windows_platform with
    | true -> ".exe"
    | false -> ""
    in
    let console =  godi_dir // "sbin" // ("godi_console" ^ exe_suffix) in
    if Sys.file_exists console then
      console
    else
      ("godi_console" ^ exe_suffix)

  let config_success_message = "<span color=\"green\"><b>Configuration successfull</b> :-)</span>"
  let apply_changes_success_message = "<span color=\"green\"><b>Request fulfilled</b> :-)</span>"

  let godi_binpkg_server = godi_binpkg_server ()
  let godi_binpkg = godi_binpkg ()
end


module Gl = struct
  let dummy_glue = object
    method get_packages = [| |] ;
    method get_wishes = [ ];
    method check_wishes ?force_rescan (_wl:wish list) = ignore ( force_rescan );  Plan_error "not initialized" ;
    method get_dependencies ~(options:( string * string ) list) (_s:string) = ignore (options); no_dependencies ;
    method remember_config ~(options:( string * string ) list) (_s:string) = ignore (options); ()
    method get_apply_order (_s:wish list) = Plan_error "not initialized" ;
    method get_bin_fln _bi = None ;
    method get_newer () = []
  end

  (* all packages *)
  let (pkgs:package array ref) = ref [| |]
  (* homepage mus be extracted from makefiles, therefore we don't look up each time again *)
  let (memo_hp: (int, string option) Hashtbl.t) = Hashtbl.create 10
  (* packages in pkgs are immutable, changes are saved in this hashtable *)
  let (changed_options: string SHashtbl.t SHashtbl.t) = SHashtbl.create 10
  let (to_do: to_do array ref) = ref [| |]
  let to_do_cnt = ref 0
  let (text_index_src: (int, unit) Hashtbl.t SHashtbl.t) = SHashtbl.create 10
  let (text_index_bin: (int, unit) Hashtbl.t SHashtbl.t) = SHashtbl.create 10
  let (glue: glue ref) = ref dummy_glue
  (* Hashtbl name -> index *)
  let (names_src: int SHashtbl.t) = SHashtbl.create 10
  let (names_bin: int list SHashtbl.t) = SHashtbl.create 10 (* name *)
  let (ident_bin: (binary_identifier,int) Hashtbl.t) = Hashtbl.create 10
  let (memo_known_dependencies: ((string * string) list, fdependencies) Hashtbl.t array ref) = ref [| |]

  let reset () =
    pkgs := [| |];
    Hashtbl.clear memo_hp ;
    SHashtbl.clear changed_options;
    to_do := [| |];
    to_do_cnt :=  0 ;
    SHashtbl.clear text_index_src ;
    SHashtbl.clear text_index_bin ;
    glue := dummy_glue;
    SHashtbl.clear names_src ;
    SHashtbl.clear names_bin ;
    Hashtbl.clear ident_bin ;
    memo_known_dependencies := [| |]

  let add_dependencies ~options ~dependencies app =
    let options = List.sort compare options
    and htl = !memo_known_dependencies.(app.id) in
    Hashtbl.replace htl options dependencies


  let get_changed_options app =
    try
      SHashtbl.find changed_options app.name
    with
    | Not_found -> let h = SHashtbl.create 0 in
                  SHashtbl.replace changed_options app.name h ;
                  h

  let get_cur_build_options app =
    match app.package_type with
    | PBinary _ -> []
    | PSource x ->
      let htl = get_changed_options app in
      let f accu (k,v) =
        let v = try SHashtbl.find htl k with | Not_found -> v in
        (k,v)::accu
      in
      List.rev ^$ List.fold_left f [] x.build_options

  let are_dependencies_known app =
    get_cur_build_options app |> List.sort compare |> Hashtbl.mem !memo_known_dependencies.(app.id)

end

module Ptype = struct
  include Ptype_glue

  let get_app_dir app_name =
    let b1 = match Util.split_one app_name '-' with
    | Util.First_last(x,_) -> x
    | Util.Nosplit ->
      Glib.Message.log `WARNING "dubious package name '%s'" app_name ;
      "apps" (* this should never happen *)
    in
    Const.godi_dir // "build" // b1 // app_name

  let get_binary_info app =
    match app.status with
    | Local_version x
    | Old_version x
    | Current_version x -> Some x.i_binary_info
    | Unused  -> match app.package_type with
      | PSource _ -> None
      | PBinary x -> Some x (*x.b_binary_info *)

  let get_fingerprint app =
    match get_binary_info app with
    | None -> None
    | Some x -> Some x.build_version_fingerprint

  let compare_package app1 app2 =
    let c = compare app1.name app2.name in
    if c <> 0 then
      c
    else
      compare_version app2.version app1.version
end
