module SHashtbl :
sig
  type key = string
  type 'a t = 'a Glue.SHashtbl.t
  val create : int -> 'a t
  val clear : 'a t -> unit
  val copy : 'a t -> 'a t
  val add : 'a t -> key -> 'a -> unit
  val remove : 'a t -> key -> unit
  val find : 'a t -> key -> 'a
  val find_all : 'a t -> key -> 'a list
  val replace : 'a t -> key -> 'a -> unit
  val mem : 'a t -> key -> bool
  val iter : (key -> 'a -> unit) -> 'a t -> unit
  val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  val length : 'a t -> int
end

type version = Glue.version

type binary_identifier = Glue.binary_identifier

type binary_info = Glue.binary_info = {
  build_version : version option;
  build_version_fingerprint : string;
  build_info : string option;
  b_ident : binary_identifier;
}

type install_information = Glue.install_information = {
  i_version : version;
  i_pretty_version : string;
  i_binary_info : binary_info;
  i_preserve : bool;
}

type source_package = Glue.source_package = {
  build_message : string;
  build_options : (string * string) list;
}

type package_type = Glue.package_type =
| PSource of source_package
| PBinary of binary_info

type status = Glue.status =
| Local_version of install_information
| Old_version of install_information
| Current_version of install_information
| Unused

type fdependencies = Glue.fdependencies

type package = Glue.package = {
  name : string;
  version : version;
  pretty_version : string;
  comment : string;
  description : string;
  godi_maintainer : string;
  install_message : string;
  dependencies : fdependencies;
  package_type : package_type;
  status : status;
  id : int;
}

val clean_exit : unit -> unit

type plan_to_do = Glue.plan_to_do =
| Plan_build of string
| Plan_build_dep of string
| Plan_remove of binary_identifier
| Plan_remove_dep of binary_identifier
| Plan_add of binary_identifier
| Plan_add_dep of binary_identifier

type wish = Glue.wish =
| Wish_build of string
| Wish_remove of string
| Wish_add of binary_identifier
| Wish_option of string * string * string

type plan_t = Glue.plan_t = Plan_ok of plan_to_do list | Plan_error of string

val ( |> ) : 'a -> ('a -> 'b) -> 'b
val ( ^$ ) : ('a -> 'b) -> 'a -> 'b

type to_do =
    Depending_install
  | Depending_upgrade
  | Depending_remove
  | Depending_reinstall
  | Install
  | Upgrade
  | Remove
  | Reinstall
  | Dont_touch

type glue = (<
             get_packages : package array;
             get_wishes: wish list ;
             check_wishes:  ?force_rescan: bool -> wish list -> plan_t ;
             get_dependencies: options:( string * string ) list -> string -> fdependencies;
             remember_config: options:(string * string) list -> string -> unit ;
             get_apply_order : wish list -> plan_t ;
             get_bin_fln:  binary_identifier -> string option ;
             get_newer: unit -> string list
                            >)

(* constants used at various places *)
module Const :
sig
  val buffer_len : int
  val invalid_string : string
  val to_install_color : string
  val to_install_dependent_color : string
  val to_remove_color : string
  val to_remove_dependent_color : string
  val categories : (string * string) list
  val status_cat_updatable : string
  val status_cat_installed : string
  val status_cat_not_installed : string
  val prog_name : string
  val windows_platform : bool
  val godi_dir : string
  val sub_prog_env: string array option
  val pixbuf_dir : string option
  val help_pixbuf : string -> GdkPixbuf.pixbuf
  val pixbuf_package_installed_and_updated : GdkPixbuf.pixbuf
  val pixbuf_package_available : GdkPixbuf.pixbuf
  val pixbuf_package_install : GdkPixbuf.pixbuf
  val pixbuf_package_upgrade : GdkPixbuf.pixbuf
  val pixbuf_package_remove : GdkPixbuf.pixbuf
  val pixbuf_package_installed_outdated : GdkPixbuf.pixbuf
  val pixbuf_update : GdkPixbuf.pixbuf
  val pixbuf_select_newer : GdkPixbuf.pixbuf
  val pixbuf_camel : GdkPixbuf.pixbuf
  val dont_show_source_packages : bool
  val godi_console : string
  val config_success_message : string
  val apply_changes_success_message : string
  val godi_binpkg_server : string option
  val godi_binpkg : string option
end


(* global mutable values and access functions *)
module Gl :
sig
  val pkgs : package array ref
  val memo_hp : (int, string option) Hashtbl.t
  val changed_options : string SHashtbl.t SHashtbl.t
  val to_do : to_do array ref
  val to_do_cnt : int ref
  val text_index_src : (int, unit) Hashtbl.t SHashtbl.t
  val text_index_bin : (int, unit) Hashtbl.t SHashtbl.t
  val glue : glue ref
  val names_src : int SHashtbl.t
  val names_bin : int list SHashtbl.t
  val ident_bin : (binary_identifier, int) Hashtbl.t
  val memo_known_dependencies :
    ((SHashtbl.key * string) list, fdependencies) Hashtbl.t array ref

  val reset : unit -> unit
  val add_dependencies :
    options:(SHashtbl.key * string) list ->
    dependencies:fdependencies -> package -> unit
  val get_changed_options : package -> string SHashtbl.t
  val get_cur_build_options : package -> (SHashtbl.key * string) list
  val are_dependencies_known : package -> bool
end


module Ptype :
sig
  val string_of_version : version -> string
  val string_of_binary_identifier : binary_identifier -> string
  val dependencies_to_string : fdependencies -> string
  val compare_version : version -> version -> int
  val string_to_version : string -> version
  val name_of_bi : binary_identifier -> string
  val version_compare_bi : binary_identifier -> binary_identifier -> int
  val get_app_dir : string -> string
  val get_binary_info : package -> binary_info option
  val get_fingerprint : package -> string option
  val compare_package : package -> package -> int
end
