open Common

let hp_regex = Str.regexp "^[ \t\n\x0B\x0C\r]*HOMEPAGE[ \t\n\x0B\x0C\r]*=[ \t\n\x0B\x0C\r]*\\([^ \t\n\x0B\x0C\r#]+\\)[ \t\n\x0B\x0C\r]*\\(#\\|$\\)"

let get_hp str =
  if String.length str < 17 then
    None
  else
    try
      let _i = Str.search_forward hp_regex str 0 in
      Some(Str.matched_group 1 str)
    with
    | Not_found -> None


let get_homepage (package:string) =
  match Util.split_one package '-' with
  | Util.First_last(prefix,_) ->
    let filename = Filename.concat Version.godi_dir ^$ Filename.concat "build" ^$
      Filename.concat prefix ^$ Filename.concat package "Makefile" in
    let homepage = ref None in
    (try
       let ch = open_in filename in
       (try
          while !homepage = None do
            homepage:=get_hp ^$ input_line ch
          done;
          close_in ch
        with
        | x -> close_in_noerr ch ; raise x);
       !homepage
     with
     | Sys_error _
     | End_of_file -> !homepage )
  | Util.Nosplit -> None


let get_homepage (package:int) =
  try
    Hashtbl.find Gl.memo_hp package
  with
  | Not_found
    ->
    let rem = get_homepage !Gl.pkgs.(package).name in
    Hashtbl.replace Gl.memo_hp package rem;
    rem

let get_cur_wl ?(to_do=(!Gl.to_do)) ?(changed_options=Gl.changed_options) () =
  let wishes : wish list ref = ref []
  and all_apps = !Gl.pkgs
  and add_end = ref None in
  for i = Array.length all_apps - 1 downto 0 do
    match to_do.(i) with
    | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall | Dont_touch
      -> ()
    | Install | Upgrade | Reinstall ->
      let app = all_apps.(i) in
      let nelem:wish = match app.package_type with
      | PSource _ -> Wish_build app.name
      | PBinary x -> Wish_add x.b_ident
      in
      wishes := nelem :: !wishes
    | Remove
      -> wishes := (Wish_remove all_apps.(i).name)::!wishes
  done;
  changed_options |> SHashtbl.iter ( fun name shtl_o_new ->
    shtl_o_new |> SHashtbl.iter ( fun key value ->
      wishes := ( Wish_option(name,key,value) )::!wishes ));
  (match !add_end with
  | None -> ()
  | Some x -> wishes := x::!wishes);
  List.rev !wishes


let compute_depends app =
  if Gl.are_dependencies_known app = false then (
    let options = Gl.get_cur_build_options app in
    let dependencies = !Gl.glue#get_dependencies ~options app.name in
    Gl.add_dependencies ~options ~dependencies app
  )

let check_plan ?force_rescan ?(wl=(get_cur_wl ())) () =
  !Gl.glue#check_wishes ?force_rescan wl

let use_conf_from_wl wl =
  let changed = SHashtbl.create ^$ List.length wl in
  let f = function
  | Wish_build _  | Wish_remove _ | Wish_add _ -> ()
  | Wish_option(pkg,key,value) ->
    try
      let app = !Gl.pkgs.(SHashtbl.find Gl.names_src pkg) in
      let (_,cur) = Gl.get_cur_build_options app |> List.find ( fun (k,_) -> k = key ) in
      if cur <> value then (
        SHashtbl.replace (Gl.get_changed_options app) key value;
        SHashtbl.replace changed pkg app;
      )
    with
    | Not_found -> ()
  in
  List.iter f wl ;
  let f _name app =
    if Gl.are_dependencies_known app = false then
      compute_depends app
  in
  SHashtbl.iter f changed


let apply_plan pl =
  let all_apps = !Gl.pkgs
  and to_do = !Gl.to_do in
  for i = Array.length all_apps - 1 downto 0 do
    to_do.(i) <- Dont_touch
  done;
  Gl.to_do_cnt := 0 ;
  match pl with
  | Plan_error _ -> ()
  | Plan_ok pl ->
    let l = ref pl
    and add = ref 0 in
    let help_remove what ident =
      let f i = match all_apps.(i).status with
      | Current_version _  | Local_version _ | Old_version _
        -> to_do.(i) <- what ; incr add
      | Unused -> ()
      in
      (try
         f ^$ SHashtbl.find Gl.names_src ^$ Ptype.name_of_bi ident
       with
       | Not_found -> ());
      (try
         f ^$ Hashtbl.find Gl.ident_bin ident
       with
       | Not_found -> ());
    in
    while !l <> [] do
      match !l with
      | [] -> assert false
      | hd::tl ->
        l := tl;
        add := 0;
        (try
           match hd with
           | Plan_build name ->
             let i = SHashtbl.find Gl.names_src name in
             to_do.(i) <- (match all_apps.(i).status with
             | Current_version _
             | Local_version _ -> Reinstall
             | Old_version _ -> Upgrade
             | Unused -> Install );
             incr add;
           | Plan_build_dep name ->
             let i = SHashtbl.find Gl.names_src name in
             to_do.(i) <- (match all_apps.(i).status with
             | Current_version _
             | Local_version _ -> Depending_reinstall
             | Old_version _ -> Depending_upgrade
             | Unused -> Depending_install );
             incr add;
           | Plan_remove ident -> help_remove Remove ident
           | Plan_remove_dep name -> help_remove Depending_remove name
           | Plan_add ident ->
             to_do.(Hashtbl.find Gl.ident_bin ident) <- Install;
             incr add;
           | Plan_add_dep ident ->
             to_do.(Hashtbl.find Gl.ident_bin ident) <- Depending_install;
             incr add;
         with
         | Not_found -> ());
        Gl.to_do_cnt := !Gl.to_do_cnt + !add ;
    done

(*
let find_binary_package_filename ~name ~version ~fprint =
  let all_apps = !Gl.pkgs
  and elem = ref None
  and i = ref 0 in
  let len = Array.length all_apps in
  while !elem = None && !i < len do
    let cur = all_apps.(!i) in
    incr i;
    match cur.package_type with
    | PSource _ -> ()
    | PBinary x ->
      let b = x.b_binary_info in
      if name = cur.name && b.build_version_fingerprint = fprint &&
        string_to_version version = b.build_version then
        elem := Some x.b_filename
  done;
  !elem
*)


let get_build_packages wl =
  let f accu = function
  | Wish_build x -> x::accu
  | Wish_remove _
  | Wish_option _
  | Wish_add _ -> accu
  in
  List.fold_left f [] wl

let get_remove_packages wl =
  let f accu = function
  | Wish_remove x -> x::accu
  | Wish_build _
  | Wish_option _
  | Wish_add _ -> accu
  in
  List.fold_left f [] wl

let get_add_packages wl =
  let f accu = function
  | Wish_add x -> x::accu
  | Wish_remove _
  | Wish_build _
  | Wish_option _ -> accu
  in
  List.fold_left f [] wl


type cpl = [
|`Name of string
|`Identifier of binary_identifier
]


let get_binary_packages_with_same_name app =
  let name = app.name
  and ident = match app.package_type with
  | PSource _ -> assert false
  | PBinary x -> x.b_ident
  in
  let f accu i =
    let app = !Gl.pkgs.(i) in
    match app.package_type with
    | PSource _ -> accu
    |  _ when app.name <> name -> accu
    | PBinary x when x.b_ident = ident -> accu
    | PBinary x -> (i,x.b_ident)::accu
  in
  let l =
    try SHashtbl.find Gl.names_bin name
    with | Not_found -> []
  in
  List.fold_left f [] l



type cpl_error = {
  not_removable: package list;
  not_installable: package list;
}

type cpl_new_consequences = {
  depending_builds: package list;
  depending_adds: package list;
  depending_removals: package list;
  updated_source_packages: package list;
  downgraded_source_packages: package list;
  rebuilded_source_packages: package list;
  updated_binary_packages: package list;
  downgraded_binary_packages: package list;
  replaced_binary_packages: package list;
}

type cpl_new =
| Cpl_ok
| Cpl_error of cpl_error
| Cpl_new_consequences of cpl_new_consequences


(*
  look after conflicts between the user wishes and the update plan generated by godi.
  to_do: to_do list before pl have been applied
  pl: return of check_plan
*)
let check_plan_local_new ~to_do ~pl =
  let not_removable = ref []
  and not_installable = ref []
  and depending_builds = ref []
  and depending_adds = ref []
  and depending_removals = ref []
  and updated_source_packages = ref []
  and downgraded_source_packages = ref []
  and rebuilded_source_packages = ref []
  and replaced_binary_packages = ref []
  and updated_binary_packages = ref []
  and downgraded_binary_packages = ref []
  and case_handled = Hashtbl.create ^$ List.length pl
  and all_apps = !Gl.pkgs in
  let add n i =
    if Hashtbl.mem case_handled i = false then (
      Hashtbl.add case_handled i ();
      n := all_apps.(i)::!n
    )
  in
  for i = 0 to Array.length to_do - 1 do
    match to_do.(i) with
    | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall | Dont_touch
      -> ()
    | Install | Upgrade | Reinstall ->
      let app = all_apps.(i) in
      (match app.package_type with
      | PSource _ ->
        let f = function
        | Plan_build x when x = app.name -> true
        | _ -> false
        in
        if List.exists f pl = false then
          add not_installable i;
      | PBinary b ->
        let f = function
        | Plan_add x when x = b.b_ident -> true
        | _ -> false
        in
        if List.exists f pl = false then
          add not_installable i
      );
    | Remove ->
      begin match all_apps.(i).package_type with
      | PSource _ -> () (* case handled by corresponding binary package *)
      | PBinary x ->
        let ident = x.b_ident in
        let f = function
        | Plan_remove x when x = ident -> true
        | _ -> false
        in
        if List.exists f pl = false then
          add not_removable i
      end
  done;
  let sort l = List.sort Ptype.compare_package l in
  if !not_removable <> [] || !not_installable <> [] then
    Cpl_error({
      not_removable =  sort !not_removable;
      not_installable = sort !not_installable})
  else
    (* order is important because of case_handled Hashtbl.
       f_build first
       f_remove second
       f_add last *)
    let ignore_remove_binary_packages = Hashtbl.create 2 in
    let f_build = function
    | Plan_build_dep name
    | Plan_build name ->
      (* ignore binary packages that will be removed, because we wan't to (re-)build a package *)
      let f i =
        match !Gl.pkgs.(i).status with
        | Unused -> ()
        | Current_version _
        | Old_version _
        | Local_version _ -> Hashtbl.replace ignore_remove_binary_packages i ()
      in
      (try SHashtbl.find Gl.names_bin name |> List.iter f with | Not_found -> () );
      (try
         let i = SHashtbl.find Gl.names_src name in
         if Hashtbl.mem case_handled i = false then
           match to_do.(i) with
           | Depending_upgrade | Depending_install | Depending_reinstall | Install | Upgrade | Reinstall -> ()
           | Remove | Depending_remove | Dont_touch ->
             match all_apps.(i).status with
             | Unused -> add depending_builds i
             | Current_version _ -> add rebuilded_source_packages i;
             | Old_version _  -> add updated_source_packages i;
             | Local_version _ -> add downgraded_source_packages i;
       with
       | Not_found -> () )
    | Plan_remove  _
    | Plan_remove_dep _
    | Plan_add_dep _
    | Plan_add _  -> ()
    and f_remove = function
    | Plan_remove ident
    | Plan_remove_dep ident ->
      (try
         let i = Hashtbl.find Gl.ident_bin ident in
         if Hashtbl.mem ignore_remove_binary_packages i = false &&
           Hashtbl.mem case_handled i = false
         then
           match to_do.(i) with
           | Remove | Depending_remove -> () (* consequences already known *)
           | Depending_upgrade | Depending_install | Depending_reinstall | Install
           | Upgrade | Reinstall | Dont_touch ->
             let cur = all_apps.(i) in
             let remove = match get_binary_packages_with_same_name cur with
             | [] -> true
             | l ->
               match List.filter ( fun (i,_) -> Hashtbl.mem case_handled i = false ) l with
               | [] -> false
               | l ->
                 let f (_,ident) =
                   let f = function
                   | Plan_add_dep s  | Plan_add s
                       when s = ident -> true
                   | Plan_build _    | Plan_build_dep _
                   | Plan_remove  _  | Plan_remove_dep _
                   | Plan_add_dep _  | Plan_add _  -> false
                   in
                   List.exists f pl
                 in
                 try
                   let (oi,other_ident) = List.find f l in
                   let cur_ident = match cur.package_type with
                   | PSource _ -> assert false
                   | PBinary x -> x.b_ident
                   in
                   let c = Ptype.version_compare_bi cur_ident other_ident in
                   if c = 0 then
                     add replaced_binary_packages i
                   else if c < 0 then
                     add updated_binary_packages i
                   else
                     add downgraded_binary_packages i;
                   Hashtbl.add case_handled oi ();
                   false
                 with
                 | Not_found -> true
             in
             if remove then
               add depending_removals i;
       with
       | Not_found -> ())
    | Plan_build_dep _
    | Plan_build _
    | Plan_add_dep _
    | Plan_add _  -> ()
    and f_add = function
    | Plan_add_dep ident
    | Plan_add ident  ->
      (try
         let i = Hashtbl.find Gl.ident_bin ident in
         if Hashtbl.mem case_handled i = false then
           begin match to_do.(i) with
           | Depending_reinstall | Depending_install | Depending_upgrade  | Install | Upgrade | Reinstall -> ()
           | Remove | Depending_remove | Dont_touch  ->
             add depending_adds i ;
           end
       with
       | Not_found -> ())
    | Plan_build_dep _
    | Plan_build _
    | Plan_remove  _
    | Plan_remove_dep _ -> ()
    in
    List.iter f_build pl;
    List.iter f_remove pl;
    List.iter f_add pl;
    if !depending_builds <> [] ||
      !depending_adds <> [] ||
      !depending_removals <> [] ||
      !replaced_binary_packages <> [] ||
      !updated_binary_packages <> [] ||
      !downgraded_binary_packages <> [] then
      Cpl_new_consequences({
        depending_builds = sort !depending_builds;
        depending_adds = sort !depending_adds;
        depending_removals = sort !depending_removals;

        rebuilded_source_packages = sort !rebuilded_source_packages;
        updated_source_packages = sort !updated_source_packages;
        downgraded_source_packages = sort !downgraded_source_packages;

        replaced_binary_packages = sort !replaced_binary_packages;
        updated_binary_packages = sort !updated_binary_packages;
        downgraded_binary_packages = sort !downgraded_binary_packages
      })
    else
      Cpl_ok


let check_plan_local ~to_do ~wl ~pl =
  let build_packages = get_build_packages wl
  and remove_packages = get_remove_packages wl
  and add_packages = get_add_packages wl in
  let  f (accu:(cpl * string) list) = function
  | Plan_build_dep name
  | Plan_build name ->
    (try
       match to_do.(SHashtbl.find Gl.names_src name) with
       | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall | Install
       | Upgrade | Reinstall | Dont_touch -> accu
       | Remove -> (`Name(name),Printf.sprintf "package '%s' was marked for removing, but it is still needed" name)::accu
     with
     | Not_found -> accu)
  | Plan_remove ident
  | Plan_remove_dep ident ->
    let f accu i = match to_do.(i) with
    | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall | Dont_touch | Remove ->
      accu
    | Install | Upgrade | Reinstall ->
      let name = Ptype.name_of_bi ident in
      (`Name name,Printf.sprintf "package '%s' was marked for installing, but it will be removed" name)::accu
    in
    let naccu =
      try f accu ^$ SHashtbl.find Gl.names_src ^$ Ptype.name_of_bi ident
      with | Not_found -> accu
    in
    if naccu != accu then
      naccu
    else (
      try f naccu ^$ Hashtbl.find Gl.ident_bin ident
      with | Not_found -> accu
    )
  | Plan_add_dep ident
  | Plan_add ident ->
    try
      let i = Hashtbl.find Gl.ident_bin ident in
      match to_do.(i) with
      | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall | Install
      | Upgrade | Reinstall | Dont_touch -> accu
      | Remove ->
        let name = !Gl.pkgs.(i).name in
        (`Identifier(ident),Printf.sprintf "package '%s' was marked for removing, but it is still needed" name)::accu
    with
    | Not_found -> accu
  in
  match List.fold_left f [] pl with
  | (_::_) as l -> List.rev l
  | [] ->
    let f_build elem = function
    | Plan_build name -> name = elem
    | Plan_build_dep _ | Plan_remove _  | Plan_remove_dep _  | Plan_add _ | Plan_add_dep _ -> false
    and f_remove elem = function
    | Plan_remove name -> Ptype.name_of_bi name = elem
    | Plan_build_dep _ | Plan_build _  | Plan_remove_dep _  | Plan_add _ | Plan_add_dep _ -> false
    and f_add ident = function
    | Plan_add x -> ident = x
    | Plan_build _ | Plan_build_dep _ | Plan_remove _  | Plan_remove_dep _  | Plan_add_dep _ -> false
    in
    let f_build accu pkg = match List.exists (f_build pkg) pl with
    | true -> accu
    | false -> (`Name pkg,Printf.sprintf "package '%s' was marked for building, but it won't be built" pkg)::accu
    and f_remove accu pkg = match List.exists (f_remove pkg) pl with
    | true -> accu
    | false ->
      let name = pkg in
      (`Name pkg,Printf.sprintf "package '%s' was marked for removing, but it won't be removed" name)::accu
    and f_add accu pkg = match List.exists (f_add pkg) pl with
    | true -> accu
    | false ->
      let name = Ptype.name_of_bi pkg in
      (`Identifier pkg,
       Printf.sprintf "package '%s' was marked for installing, but it can't be installed" name
      )::accu
    in
    let accu = List.fold_left f_add [] add_packages in
    let accu = List.fold_left f_build accu build_packages in
    List.rev ^$ List.fold_left f_remove accu remove_packages




let strs l = List.map ( fun a -> a.name ) l
let show_cpl_error_dialog er =
  assert ( er.not_removable <> [] || er.not_installable <> [] );
  let tree =
    ("Not removable:", true, strs er.not_removable)
    ::("Not installable:",true,strs er.not_installable)
    ::[]
  and window_title = "Action not possible"
  and title = "Action not possible"
  and info = "Sorry, I can't proceed!\nYour wishes can't be fullfilled :-("
  in
  Info_tree.show_error ~tree ~window_title ~title ~info ()

let ask_cpl_consequences cq =
  let tree =
    ("To be removed:",true, strs cq.depending_removals)

    ::("To be downgraded:",true,strs cq.downgraded_binary_packages)
    ::("Rebuilding older source packages:",true, strs cq.downgraded_source_packages)

    ::("Rebuilding source packages:",false,strs cq.rebuilded_source_packages)
    ::("To be replace by other builds:",false,strs cq.replaced_binary_packages)

    ::("Building newer versions:",false,strs cq.updated_source_packages)
    ::("Installing newer versions:",false,strs cq.updated_binary_packages)

    ::("To be built",false,strs cq.depending_builds)
    ::("To be added",false,strs cq.depending_adds)
    ::[]
  and window_title = "Proceed anyway?"
  and title = "Mark additional required changes?"
  and info = "The chosen action also affects other packages.\nThe following changes are required in order to proceed:"
  in
  Info_tree.ask ~tree ~window_title ~title ~info ()

let try_apply_to_do to_do =
  match check_plan ~wl:(get_cur_wl ~to_do ()) () with
  | (Plan_ok pl) as pl_whole  ->
    let do_apply_plan = match check_plan_local_new ~to_do ~pl with
    | Cpl_ok  -> true
    | Cpl_new_consequences cql ->
      ask_cpl_consequences cql
    | Cpl_error el ->
      show_cpl_error_dialog el;
      false
    in
    begin match do_apply_plan with
    | false -> false
    | true ->
      apply_plan pl_whole;
      let active = !Gl.to_do_cnt > 0 in
      Toolbar.apply_button#misc#set_sensitive active;
      Toolbar.cancel_button#misc#set_sensitive active;
      true
    end
  | Plan_error(message) ->
    let () = Info_simple.show_error ~window_title:"Mission possible" ~title:"Mission impossible" ~message () in
    false
