open Pipe

module SHashtbl = Hashtbl.Make(struct
  type t = string
  let equal (a:string) (b:string) = a = b
  let hash = Hashtbl.hash
end)


type version = string * int


type fdependencies =  Godi_versions.dependency list * Godi_versions.dependency list

(* name * version * fprint *)
type binary_identifier = string * string * string * fdependencies

type binary_info = {
  build_version : version option;
  build_version_fingerprint : string ;
  build_info : string option;
  b_ident: binary_identifier ;
}

type install_information = {
  i_version: version;
  i_pretty_version: string;
  i_binary_info: binary_info;
  i_preserve: bool;
}

type source_package = {
  build_message : string ;
  build_options : (string * string) list ;
}

(*
type binary_package = {
  b_binary_info: binary_info;
  }
*)

type package_type =
| PSource of source_package
| PBinary of binary_info

type status =
| Local_version of install_information (* newer than avaiable, should not happen *)
| Old_version of install_information (* updatable *)
| Current_version of install_information (* installed in current version *)
| Unused

type package = {
  name : string;
  version : version ;
  pretty_version: string;
  comment : string ;
  description : string ;
  godi_maintainer : string ;
  install_message : string ;
  dependencies : fdependencies ;
  package_type: package_type;
  status: status;
  id: int;
}


type plan_to_do =
| Plan_build of string    (* package name *)
| Plan_build_dep of string (* package name *)
| Plan_remove of binary_identifier   (* package name *)
| Plan_remove_dep of binary_identifier (* package name *)
| Plan_add of binary_identifier  (* identifier *)
| Plan_add_dep of binary_identifier (* identifier *)

type wish =
| Wish_build of string
| Wish_remove of string
| Wish_add of binary_identifier
| Wish_option of string * string * string


type plan_t =
| Plan_ok of plan_to_do list
| Plan_error of string



let () =
  Godi_exec.set_variable "OCAMLFIND_DESTDIR" "";
  Godi_exec.set_variable "OCAMLFIND_METADIR" "";
  Godi_exec.set_variable "OCAMLPATH" "";
  Godi_exec.set_variable "OCAMLFIND_COMMANDS" "";
  Godi_exec.set_variable "OCAMLFIND_LDCONF" ""


(*
let version_to_pretty_string (version,revision) =
  if revision > 0 then
    version ^ "#" ^ string_of_int revision
  else
    version

let version_from_package pkg =
  let x = Godi_types.any_pkg pkg in
  (x#version,x#pkgrevision)
*)


let con = Godi_types.stderr_console()

let clean_exit = Godi_file.release_all_tmpdirs

(* taken from cmd.ml *)

module Ptype_glue = struct
  let string_of_version (v,r) =
    if r > 0 then v ^ "#" ^ string_of_int r else v

  let string_of_binary_identifier (name,version,fprint,_) = Printf.sprintf "%s-%s@%s" name version fprint

  let dependencies_to_string (real,build) =
    let real = List.map Godi_versions.print_dependency real
    and build =  List.map (fun a -> (Godi_versions.print_dependency a) ^ " (build only)" ) build in
    String.concat "\n" (real@build)


  let compare_version (v1,r1) (v2,r2) =
    match Godi_versions.cmp_version_string v1 v2 with
    | 0 ->
      if r1 < 0 || r2 < 0 then
        0
      else
        compare r1 r2
    | x -> x


  let string_to_version x = (x,(-1))

  let name_of_bi (s,_,_,_) = s

  let version_compare_bi (n1,v1,_,_) (n2,v2,_,_) =
    assert ( n1 = n2 );
    Godi_versions.cmp_version_string v1 v2

end
open Ptype_glue

let version_str_any pkg =
  string_of_version (pkg#version,pkg#pkgrevision)

(*let version_str pkg =
  version_str_any (Godi_types.any_pkg pkg) *)





let remove_linefeeds s =
  Str.global_replace (Str.regexp "[\r\n]") " " s

(* use x#variant vid? *)
let copy_elem = function
| `Source_pkg s -> `Source_pkg (Oo.copy s)
| `Binary_pkg s -> `Binary_pkg (Oo.copy s)
| `Installed_pkg s -> `Installed_pkg (Oo.copy s)
| `Dispersed_pkg s -> `Dispersed_pkg (Oo.copy s)
| `Wrapped_pkg s -> `Wrapped_pkg (Oo.copy s)


let update_src_pkgs ?(force_rescan=false) ?conf_file src_pkgs =
  let tmp_godi_conf,tmp_dir = match conf_file with
  | Some x -> x, None
  | None ->
      (* Write a temporary godi.conf file: *)
    let tmp = Godi_file.allocate_tmpdir() in
    let tmp_path = Godi_file.tmpdir_path tmp in
    Godi_file.addname tmp_path "godi.conf", Some tmp
  in

  let mflag = Godi_conffile.modify_godi_conf ~destfile:tmp_godi_conf src_pkgs in

  (* When options have been modified, rescan source packages: *)
  let src_pkgs =
    if not mflag && force_rescan = false then
      src_pkgs
    else
      let vid = Godi_types.new_var_id() in
      List.iter (fun pkg ->
        (Godi_types.source_pkg pkg)#rescan_deps con (Godi_file.slashname tmp_godi_conf) vid
      ) src_pkgs;
      List.map
	(fun pkg -> `Source_pkg ((Godi_types.source_pkg pkg)#variant vid))
	src_pkgs
  in
  (match tmp_dir with
  | None -> ()
  | Some x -> Godi_file.release_tmpdir x);
  src_pkgs


(*let filename_binary pkg =
  try
    Godi_file.slashname (Godi_types.binary_pkg pkg)#binary_archive
  with
  | Failure _ -> "" *)


let get_godi_var name =
  try
    Some(Godi_query.from_godi_conf name)
  with
  | _ -> None




let get_binary_ident = function
| `Binary_pkg p -> p#name,p#version,p#build_version_fingerprint,(p#dependencies,p#build_dependencies)
| `Installed_pkg p -> p#name,p#version,p#build_version_fingerprint,(p#dependencies,p#build_dependencies)
| _ -> failwith "Glue.get_binary_ident: package has bad type "


let scan_all_binary_packages con =
  let htl = Hashtbl.create 3 in
  let scan_binary_packages url =
    let handler = Godi_file_handler.handler_of_directory url in
    let pkgmanifest, rel = handler # get con "PKGMANIFEST"  in
    let pkgman_list = Godi_query.split_manifest pkgmanifest in
    ( match rel with
    | None -> ()
    | Some f -> f()
    );
    List.map
      (fun inst_plist ->
        let file = Godi_plist.extract_binpkg inst_plist in
        let pkg = Godi_query.get_remote_binary_package
          inst_plist
          (url ^ "/" ^ file)   (* FIXME: incorrect (URL encoding) *)
        in
        Hashtbl.replace htl (get_binary_ident pkg) file ;
        pkg
      )
      pkgman_list
  in
  let p = Godi_query.godi_binpkg_path() in
  htl,List.flatten (List.map  scan_binary_packages p )


let transform_filename name =
  Godi_file.filename name |> Godi_file.nativename

let godi_binpkg_server_lazy = lazy ( get_godi_var "GODI_BINPKG_SERVER" )
let godi_binpkg_lazy = lazy (
  match get_godi_var "GODI_BINPKG" with
  | None -> None
  | Some x ->
    match Util.remove_spaces_begin_end x with
    | "" -> None
    | x -> try Some (transform_filename x ) with _ -> None
)

let godi_binpkg_server () = Lazy.force godi_binpkg_server_lazy
let godi_binpkg () = Lazy.force godi_binpkg_lazy

(*let get_fingerprint = function
| `Binary_pkg p -> p#build_version_fingerprint
| `Installed_pkg p -> p#build_version_fingerprint
| _ -> "" *)

let deconstructor tmp_dir =
  match Lazy.lazy_is_val tmp_dir with
  | false -> ()
  | true -> Godi_file.release_tmpdir ^$ Lazy.force tmp_dir


let get_binary_ident_unboxed p = p#name,p#version,p#build_version_fingerprint,(p#dependencies,p#build_dependencies)

let to_godi_wl wl =
  let f = function
  | Wish_build s -> `Build s
  | Wish_add(s1,s2,s3,_) -> `Add(s1,s2,s3)
  | Wish_remove s -> `Remove s
  | Wish_option(s1,s2,s3) -> `Option(s1,s2,s3)
  in
  List.map f wl


class ginfo ?(no_source=false) () =
  let inst_pkgs =
    Godi_query.notify_about_change_of_installed_packages con `Reload ;
    Godi_query.get_installed_packages con
  in
  let htl_fln_bin_pkgs,bin_pkgs = scan_all_binary_packages con in
object(self)
  val mutable src_pkgs =   match no_source with
  | true -> []
  | false ->  Godi_query.forget_source_packages (); Godi_query.scan_source_packages con
  val shtl_options = SHashtbl.create 150
  val tmp_dir = Lazy.lazy_from_fun Godi_file.allocate_tmpdir
  val mutable conf_file = None
  val mutable conf_changed = false


  initializer
  let f pkg =
    let l = (Godi_types.source_pkg pkg)#values_of_build_options
    and f htl (a,v) = SHashtbl.replace htl a v ; htl in
    let htl = SHashtbl.create ^$ List.length l in
    ignore ( List.fold_left f htl l );
    SHashtbl.replace shtl_options (Godi_types.pkg_name pkg) htl
  in
  List.iter f src_pkgs;
  Gc.finalise deconstructor tmp_dir

  method private conf_file =
    match conf_file with
    | Some x -> x
    | None ->
      let tmp_path = Godi_file.tmpdir_path ^$ Lazy.force tmp_dir  in
      let tmp_godi_conf = Godi_file.addname tmp_path "godi.conf" in
      conf_file <- Some tmp_godi_conf;
      tmp_godi_conf


  method get_bin_fln bident =
    try
      Some(Hashtbl.find htl_fln_bin_pkgs bident)
    with
    | _ -> None

  method get_wishes =
    let shtl_options = SHashtbl.create 2
    and shtl_build = SHashtbl.create 2
    and shtl_remove = SHashtbl.create 2 in
    let module U = struct exception Unsupported end in

    let wishes = Godi_wishes.read_wishes ~pref_user_error:true () in

    (*sanitize wish list, only simple wish lists are supported *)
    let f = function
    | `Build(s) ->
      if SHashtbl.mem shtl_build s || SHashtbl.mem shtl_remove s then raise U.Unsupported;
      SHashtbl.add shtl_build s ()
    | `Remove(s) ->
      if SHashtbl.mem shtl_build s || SHashtbl.mem shtl_remove s then raise U.Unsupported;
      SHashtbl.add shtl_remove s ()
    | `Add _ -> () (* not yet supported *)
    | `Option(pkg,key,value) ->
      let h = try SHashtbl.find shtl_options pkg with
      | Not_found ->
        let x = SHashtbl.create 1 in
        SHashtbl.add shtl_options pkg x ;
        x
      in
      if SHashtbl.mem h key then
        raise U.Unsupported;
      SHashtbl.replace h key value
    in

    try
      List.iter f wishes;
      let wishes = SHashtbl.fold ( fun key () accu -> (Wish_build key)::accu ) shtl_build [] in
      let wishes = ref ( SHashtbl.fold ( fun key () accu -> (Wish_remove key)::accu ) shtl_remove wishes ) in
      shtl_options |> SHashtbl.iter ( fun name shtl_o_new ->
        let shtl_o_orig = SHashtbl.find shtl_options name in
        shtl_o_new |> SHashtbl.iter ( fun key value ->
          let orig_value = try SHashtbl.find shtl_o_orig key with | Not_found -> "\000\000" in
          if orig_value <> value then
            wishes := ( Wish_option(name,key,value) )::!wishes ));
      List.rev !wishes
    with
    | U.Unsupported -> []


  method get_apply_order wishes =
    try
      let wishes = to_godi_wl wishes in
      Godi_wishes.apply_wishes_to_packages src_pkgs wishes;
      Godi_wishes.apply_wishes_to_packages inst_pkgs wishes;
      Godi_wishes.apply_wishes_to_packages bin_pkgs wishes;
      let _n = Godi_plan.update_plan ~iteration:1 (src_pkgs @ inst_pkgs @ bin_pkgs) in
      let to_remove = List.rev ^$ Godi_plan.sort_packages ^$ List.filter Godi_types.to_be_removed inst_pkgs in
      let to_build_or_add = Godi_plan.sort_packages (
        List.filter Godi_types.to_be_built src_pkgs @ List.filter Godi_types.to_be_added bin_pkgs
      )
      in
      let f_remove (accu:plan_to_do list) pkg =
        let nelem = match Godi_types.pkg_user_status pkg with
        | `Remove_dep _ -> Plan_remove_dep(get_binary_ident pkg)
        | `Remove -> Plan_remove(get_binary_ident pkg)
        | `Keep
        | `Add
        | `Build
        | `Add_dep _
        | `Build_dep _ -> assert false
        in
        nelem::accu
      and f_add accu pkg =
        let apkg = Godi_types.any_pkg pkg in
        let name = apkg#name in
        let nelem = match apkg#user_status with
        | `Remove_dep _
        | `Keep
        | `Remove -> assert false
        | `Add -> Plan_add(get_binary_ident pkg);
        | `Add_dep _ -> Plan_add_dep(get_binary_ident pkg)
        | `Build  -> Plan_build(name)
        | `Build_dep _ -> Plan_build(name)
        in
        nelem::accu
      in
      let accu = List.fold_left f_remove [] to_remove in
      let accu = List.fold_left f_add accu to_build_or_add in
      let erg = List.rev accu in
      self#save_config to_build_or_add ;
      Plan_ok erg
    with
    | Godi_plan.Mission_impossible text ->
      Plan_error text
    | Godi_plan.Package_installation_not_unique name ->
      Plan_error("Tried to install several versions of this package at once: " ^ name)



  method private save_config pkgs =
    if conf_changed = true && pkgs <> [] then
      ignore ( Godi_conffile.modify_godi_conf pkgs )


  method remember_config ~options package =
    let shtl_o_orig = SHashtbl.find shtl_options package in
    let wishes : Godi_wishes.wish list = List.fold_left ( fun accu (key,value) ->
      let orig_value = try SHashtbl.find shtl_o_orig key with | Not_found -> "\000\000" in
      if orig_value = value then
        accu
      else (
        SHashtbl.replace shtl_o_orig key value ;
        (`Option(package,key,value))::accu
      )
    ) [] options  |> List.rev
    in
    match wishes with
    | [] -> ()
    | _ ->
      Godi_wishes.apply_wishes_to_packages src_pkgs wishes;
      conf_changed <- true;
      src_pkgs <- update_src_pkgs ~force_rescan:true ~conf_file:self#conf_file src_pkgs;



  method get_dependencies ~options package =
    let shtl_o_orig = SHashtbl.find shtl_options package in
    let wishes : Godi_wishes.wish list = List.fold_left ( fun accu (key,value) ->
      let orig_value = try SHashtbl.find shtl_o_orig key with | Not_found -> "\000\000" in
      if orig_value <> value then
        (`Option(package,key,value))::accu
      else
        accu
    ) [] options  |> List.rev
    in

    let src_pkgs' = match wishes with
    | [] -> src_pkgs
    | _ ->
      let src_pkgs = List.map copy_elem src_pkgs in
      Godi_wishes.apply_wishes_to_packages src_pkgs wishes;
      update_src_pkgs src_pkgs
    in
    let f pkg =
      match pkg with
      | `Source_pkg src -> src#name = package
      | _ -> false
    in
    match List.find f src_pkgs' with
    | `Source_pkg src -> (src#dependencies,src#build_dependencies)
    | _ -> assert false


  method get_newer () =
    let src_pkgs_map = Godi_types.PkgMap.make (List.map (fun pkg -> (pkg,())) src_pkgs) in
    let f accu pkg =
      let name = Godi_types.pkg_name pkg in
      let spkgs = Godi_types.PkgMap.find_by ~name src_pkgs_map in
      List.fold_left (fun accu (spkg,_) ->
	if Godi_types.newer_package spkg pkg then
          (Godi_types.pkg_name spkg)::accu
        else
          accu
      ) accu spkgs
    in
    List.fold_left f [] inst_pkgs


  method check_wishes ?force_rescan wishes =
    let src_pkgs = List.map copy_elem src_pkgs
    and inst_pkgs = List.map copy_elem inst_pkgs
    and bin_pkgs = List.map copy_elem bin_pkgs
    and wishes = to_godi_wl wishes in
    let config_changed  = function
    | `Option(package,key,value) ->
      let shtl_o_orig = SHashtbl.find shtl_options package in
      let orig_value = try SHashtbl.find shtl_o_orig key with | Not_found -> "\000\000" in
      orig_value <> value
    | _ -> false
    in
    Godi_wishes.apply_wishes_to_packages src_pkgs wishes;
    Godi_wishes.apply_wishes_to_packages inst_pkgs wishes;
    Godi_wishes.apply_wishes_to_packages bin_pkgs wishes;
    let src_pkgs = match List.exists config_changed wishes with
    | false ->  src_pkgs
    | true -> update_src_pkgs ?force_rescan src_pkgs
    in
    try
      let _n = Godi_plan.update_plan ~replace:`Upgrade ~iteration:1  (src_pkgs @ inst_pkgs @ bin_pkgs ) in

      let f (accu:plan_to_do list) pkg =
        let user_status = (Godi_types.any_pkg pkg)#user_status in
        match user_status with
        | `Keep -> accu
        | _ ->
          let name = (Godi_types.pkg_name pkg) in
          let nelem = match user_status with
            | `Keep -> assert false;
            | `Add  -> Plan_add(get_binary_ident pkg)
            | `Add_dep _ ->Plan_add_dep(get_binary_ident pkg)
            | `Build  -> Plan_build name
            | `Remove -> Plan_remove (get_binary_ident pkg)
            | `Build_dep _ -> Plan_build_dep name
            | `Remove_dep _ -> Plan_remove_dep (get_binary_ident pkg)
            in
            nelem::accu
      in
      let accu = List.fold_left f [] src_pkgs in
      let accu = List.fold_left f accu bin_pkgs in
      Plan_ok( Util.list_unique ^$ List.fold_left f accu inst_pkgs )
    with
    | Godi_plan.Mission_impossible text ->
      Plan_error text
    | Godi_plan.Package_installation_not_unique name ->
      Plan_error("Tried to install several versions of this package at once: " ^ name)


  method get_packages =

    let h_installed =
      let f  h_installed  = function
      | `Source_pkg _
      | `Dispersed_pkg _
      | `Wrapped_pkg _
      | `Binary_pkg _ -> assert false
      | `Installed_pkg pkg ->
        Hashtbl.replace h_installed pkg#name pkg;
        h_installed
      in
      let h = Hashtbl.create ^$ List.length inst_pkgs in
      List.fold_left f h inst_pkgs
    in

    let get_installed name =
      try
        Some( Hashtbl.find h_installed name )
      with
        Not_found -> None
    in

    let f_common b_unbox package_type status =
      let name = b_unbox#name
      and version = b_unbox#version,b_unbox#pkgrevision
      and pretty_version = version_str_any (b_unbox :> Godi_types.any_package )
      and comment = remove_linefeeds b_unbox#comment
      and description = b_unbox#description
      and godi_maintainer =  b_unbox#godi_maintainer
      and install_message = "" (* why this exception??? *)
      and dependencies = b_unbox#dependencies,b_unbox#build_dependencies
      and id = 0 in
      {name; version ; pretty_version ; comment ; description ; godi_maintainer ;
       install_message ; dependencies ; package_type ; status; id }
    in

    let f_binary_info pkg = {
      build_version = (try Some(pkg#build_version,(-1)) with | Failure _ -> None);
      build_version_fingerprint = pkg#build_version_fingerprint;
      build_info = (try Some(pkg#build_info) with | Failure _ -> None );
      b_ident = get_binary_ident_unboxed pkg
    }
    in

    let f_inst_info pkg =
      {
        i_version= pkg#version,pkg#pkgrevision;
        i_pretty_version= version_str_any (pkg :> Godi_types.any_package );
        i_binary_info= f_binary_info pkg;
        i_preserve = pkg#preserve ;
      }
    in

    let f_installed _name inst_pkg accu =
      let status = Current_version(f_inst_info inst_pkg)
      and package_type = PBinary(f_binary_info inst_pkg) in
      (*and package_type = PBinary( {
        b_binary_info = f_binary_info inst_pkg;
        b_ident = get_binary_ident ^$ `Installed_pkg inst_pkg
      }) in *)
      (f_common inst_pkg package_type status)::accu
    in

    let f_binary accu (x:Godi_types.package) = match x with
    | `Source_pkg _
    | `Installed_pkg _ -> assert false
    | `Dispersed_pkg _
    | `Wrapped_pkg _ -> accu
    | `Binary_pkg b_unbox ->
      let name = b_unbox#name in
      let status = match get_installed name with
      | None -> Unused
      | Some x ->
        if x#version <> b_unbox#version ||
          x#build_version_fingerprint <> b_unbox#build_version_fingerprint then
          Unused
        else (
          Hashtbl.remove h_installed name;
          Current_version(f_inst_info x)
        )
      and package_type = PBinary(f_binary_info b_unbox) in
      (f_common b_unbox package_type status)::accu
    in

    let f_src accu = function
    | `Dispersed_pkg _
    | `Wrapped_pkg _  -> accu
    | `Installed_pkg _
    | `Binary_pkg _ -> assert false
    | `Source_pkg src_unboxed ->
      let inst_unboxed = get_installed src_unboxed#name in
      let status = match inst_unboxed with
      | None -> Unused
      | Some inst_unboxed  ->
        let inst_info = f_inst_info inst_unboxed in
        let c = Godi_versions.cmp_version_string src_unboxed#version inst_unboxed#version in
        let c = match c with
        | 0 -> compare src_unboxed#pkgrevision inst_unboxed#pkgrevision
        | x -> x
        in
        if c = 0 then
          Current_version inst_info
        else if c > 0 then
          Old_version inst_info
        else
          Local_version inst_info
      and package_type = PSource({
        build_message = src_unboxed#build_message ;
        build_options = src_unboxed#values_of_build_options
      })
      in
      let app = f_common src_unboxed package_type status in
      app::accu
    in
    let apps = List.fold_left f_src [] src_pkgs in
    let apps = List.fold_left f_binary apps bin_pkgs in
    let apps = Hashtbl.fold f_installed h_installed apps in
    let sort_f a b =
      match compare a.name b.name with
      | 0 -> compare_version b.version a.version
      | x -> x
    in
    let apps = List.sort sort_f apps in
    let id_counter = ref 0 in
    let f a =
      let erg = { a with id = !id_counter } in
      incr id_counter;
      erg
    in
    let apps = List.map f apps in
    Array.of_list apps
end

let no_dependencies = ([],[])
