module SHashtbl :
  sig
    type key = string
    type 'a t
    val create : int -> 'a t
    val clear : 'a t -> unit
    val copy : 'a t -> 'a t
    val add : 'a t -> key -> 'a -> unit
    val remove : 'a t -> key -> unit
    val find : 'a t -> key -> 'a
    val find_all : 'a t -> key -> 'a list
    val replace : 'a t -> key -> 'a -> unit
    val mem : 'a t -> key -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val length : 'a t -> int
  end

type version


type binary_identifier

type binary_info = {
  build_version : version option;
  build_version_fingerprint : string ;
  build_info : string option;
  b_ident: binary_identifier ;
}


type install_information = {
  i_version: version;
  i_pretty_version: string;
  i_binary_info: binary_info;
  i_preserve: bool;
}

type source_package = {
  build_message : string ;
  build_options : (string * string) list ;
}

type package_type =
| PSource of source_package
| PBinary of binary_info

type status =
| Local_version of install_information
| Old_version of install_information
| Current_version of install_information
| Unused

type fdependencies

type package = {
  name : string;
  version : version ;
  pretty_version: string;
  comment : string ;
  description : string ;
  godi_maintainer : string ;
  install_message : string ;
  dependencies : fdependencies ;
  package_type: package_type;
  status: status;
  id:int;
}

val clean_exit : unit -> unit



type plan_to_do =
| Plan_build of string    (* package name *)
| Plan_build_dep of string (* package name *)
| Plan_remove of binary_identifier   (* package name *)
| Plan_remove_dep of binary_identifier (* package name *)
| Plan_add of binary_identifier  (* identifier *)
| Plan_add_dep of binary_identifier (* identifier *)

type wish =
| Wish_build of string
| Wish_remove of string
| Wish_add of binary_identifier
| Wish_option of string * string * string

type plan_t =
| Plan_ok of plan_to_do list
| Plan_error of string

class ginfo : ?no_source:bool -> unit -> object
  (* test if applying user wishes is possible.
     force_rescan: force rescan of dependencies because of changed configurations
  *)
  method check_wishes : ?force_rescan: bool -> wish list -> plan_t

  (* apply changes in options to current source packages,
     so we don't need to recompute it all the time *)
  method remember_config: options:(string * string) list -> string -> unit

  (* compute dependencies for current options.
     Dubious background assumption: dependencies can only change after changing options of
     the current package
  *)
  method get_dependencies : options:(string * string) list -> string -> fdependencies

  method get_packages : package array

  (* try to use values of /opt/godi/build/wishes.txt *)
  method get_wishes : wish list

  (* called immediately before executing plan_t, therefore also save the current configuration
     Packages are removed with:
        godi_console delete -v package_name
     Packages are fetched with:
        godi_console make fetch
     Packages are build with:
        godi_console make clean build install package clean
     Packages are added with
        godi_console add -v package
  *)
  method get_apply_order : wish list -> plan_t

  method get_bin_fln:  binary_identifier -> string option

  (* get a list of source packages that can be updated. Binary packages are not supported yet *)
  method get_newer: unit -> string list
end

module Ptype_glue :
  sig
    val string_of_version: version -> string
    val string_of_binary_identifier : binary_identifier -> string
    val dependencies_to_string: fdependencies -> string
    val compare_version: version -> version -> int
(* still needed because of wish list *)
    val string_to_version: string -> version
    val name_of_bi: binary_identifier -> string
    val version_compare_bi: binary_identifier -> binary_identifier -> int
  end


val no_dependencies: fdependencies







val godi_binpkg_server : unit -> string option
val godi_binpkg: unit -> string option
