
type referer_option =
  | Referer_disable
  | Referer_fake
  | Referer of string

type file_exist_action =
| Overwrite
| Inform of (string -> unit)
| Stop

let default_user_agent = "Netclient"
let default_referer = Referer_fake

type error = | File_exists

exception EcurlExeception of error * string


let get_url_prefix_helper url =
  let ulen = String.length url in
  if ulen < 7 then
    raise Not_found;
  if url.[0] <> 'h' || url.[1] <> 't' || url.[2] <> 't' || url.[3] <> 'p' then
    raise Not_found;
  let i = ref (match url.[4] with
    |  's' -> 5
    | _ -> 4)
  in
  if url.[!i] <> ':' then
    raise Not_found;
  incr i;
  if url.[!i] <> '/' then
    raise Not_found;
  incr i;
  if url.[!i] <> '/' then
    raise Not_found;
  incr i;
  let found = ref false in
  while !i < ulen && !found= false do
    if url.[!i] = '/' then
      found:=true;
    incr i;
  done ;
  String.sub url 0 !i

let get_url_prefix url =
  try
    get_url_prefix_helper url
  with
    | Not_found -> url

let get_mime_type _file = None
(* disabled because I don't want a additional dependency
   try
    let m = Magic.make [] in
    try
      Magic.setflags m [ Magic.Mime ] ;
      let erg = Magic.file m file in
      Magic.close m ;
      Some erg
    with
      | _ -> Magic.close m ; None
  with
    | _ -> None
*)

let get_filename_from_url url =
  try
    let p = String.rindex url '/' in
    match String.sub url (succ p) ( (String.length url) - p - 1 ) with
      | "" -> None
      | x -> Some x
  with
    | Not_found -> None


let get_file_suffix_from_mime_type mime =
  match get_filename_from_url mime with
    | None -> "dat"
    | Some x ->
      try
	let p = String.rindex x '.' in
	match String.sub x (succ p) ( (String.length x) - p - 1 ) with
	  | "" -> "dat"
	  | x -> (String.lowercase x)
      with
	| Not_found ->(String.lowercase x)

(*
let split_one str chr =
  try
    let pos = String.index str chr in
    let first = String.sub str 0 (pos)
    and second = String.sub str (succ pos) ((String.length str) - pos - 1) in
    first,second
  with
    | Not_found -> (str,"")
      *)

let lazy_init = lazy (
  Curl.global_init Curl.CURLINIT_GLOBALALL;
  at_exit Curl.global_cleanup )


class ecurl
  ?(user_agent=default_user_agent)
  ?(use_cookies=false)
(*  ?(proxy_kind=None)
  ?(proxy_address="") *)
  ?(default_referer=default_referer)
  ()
  =
object (self)
  val cookie_file =
    let tmp=
      let () = Lazy.force lazy_init in
      match use_cookies with
	| false -> ""
	| true ->
	  let tmp_file = Filename.temp_file "cookies" ".txt" in
	  let del_fun = fun () ->
	    Sys.remove tmp_file
	  in
	  at_exit del_fun ;
	  tmp_file
    in
    tmp
(*  val proxy =
    begin match proxy_kind with
      | None -> "",(-1)
      | _ ->
	begin match split_one proxy_address ':' with
	  | "",_ -> invalid_arg "proxy address missing"
	  | add,"" -> add,(-1)
	  | add,port ->
	    begin
	      try
		add,(int_of_string port)
	      with
	      | _ -> invalid_arg "invalid port"
	    end
	end
    end *)

  val error_buffer = ref ""
  val mutable opt_con = None

  method private get_con =
    match opt_con with
      | None -> raise (Failure "error in ecurl get_con")
      | Some x -> x

  method private clean =
    match opt_con with
      | None -> ()
      | Some(x) ->
	let t = x#cleanup in
	opt_con <- None ;
	t
(*
  method private handle_curl_exception error =
    self#clean ;
    match error with
      | Curl.CurlException(_,code,str) ->
	let msg = Printf.sprintf "%s (%d,%s)" !error_buffer code str in
	raise (EcurlExeception(Curl_error,msg))
      | x -> raise x
*)

  method private curl_init ~referer url =
    let con = new Curl.handle in
    opt_con <- Some(con) ;

    con#set_errorbuffer error_buffer;
    con#set_failonerror true;
    (*begin match proxy_kind with
    | None -> ()
    | Some kind ->
      con#set_proxytype kind ;
      con#set_proxy (fst proxy);
      begin match snd proxy with
      | -1 -> ()
      |  p -> con#set_proxyport p ;
      end
    end;*)
    if cookie_file <> "" then (
      con#set_cookiejar cookie_file ;
    );
    con#set_useragent user_agent;

    begin match referer with
    | Referer_disable -> ()
    | Referer_fake -> con#set_referer (get_url_prefix url)
    | Referer(s) -> con#set_referer s
    end ;
    con#set_followlocation true;
    con#set_url url


  method private getpostcommon =
    Util.try_finally ( fun () ->
      let con = self#get_con in
      let result = Buffer.create 16384 in
      let result_helper data =
        Buffer.add_string result data;
        String.length data
      in
      con#set_writefunction result_helper;
      con#perform;
      Buffer.contents result
    ) () ( fun () -> self#clean ) ()

  method get ?(referer=default_referer)  url =
    self#curl_init ~referer url ;
    self#getpostcommon


  method post ?(referer=default_referer) ?(files=[]) params url =
    self#curl_init ~referer url ;
    let con = self#get_con in
    let postvar = ref [] in
    let helper_params = fun (name,value) ->
      postvar := (Curl.CURLFORM_CONTENT(name,value,Curl.DEFAULT))::!postvar
    and helper_files = fun (name,file) ->
      let mime =
	match get_mime_type file with
	| None -> Curl.DEFAULT
	| Some x -> Curl.CONTENTTYPE x
      in
      postvar := (Curl.CURLFORM_FILE(name,file,mime))::!postvar
    in
    List.iter helper_params  params ;
    List.iter helper_files files;

    con#set_httpheader [ "Expect:" ];
    con#set_httppost (List.rev !postvar);
    self#getpostcommon


  method private get2file_common ?fprogress ?(real_url=None) ~referer ~file url =
    Util.try_finally ( fun () ->
      self#curl_init ~referer url ;
      let con = self#get_con
      and ch = open_out_gen [  Open_wronly; Open_binary ; Open_creat ; 	Open_trunc ] 0o600 file in
      Util.try_finally ( fun () ->
	let result_helper data =
	  output_string ch data;
	  String.length data
	in
	con#set_writefunction result_helper;
	con#set_filetime true;
        (match fprogress with
        | None -> ()
        | Some x ->
          con#set_noprogress false;
          con#set_progressfunction x);
	con#perform ;
	let time = con#get_filetime in
	begin match real_url with
	| None -> ()
	| Some x -> x:= con#get_effectiveurl
	end;
	(* unimportant detail, ignore error *)
	(try Unix.utimes file time time with | _ -> ());
      ) () close_out_noerr ch
    ) () ( fun () -> self#clean ) ()


  method get2file ?(referer=default_referer) ?fprogress ~file url =
    self#get2file_common ?fprogress ~referer ~file url

  method get2dir ?(referer=default_referer) ?fprogress ?(ask=Overwrite) ~dir url =
    let file = Filename.temp_file ~temp_dir:dir "tempcurl" ".tmp" in
    let real_url = ref "" in
    self#get2file_common ?fprogress ~real_url:(Some real_url) ~referer ~file url ;
    let filename = match get_filename_from_url !real_url with
      | Some x -> x
      | None ->
	match get_mime_type file with
	  | None -> "index.dat"
	  | Some x -> "index." ^ (get_file_suffix_from_mime_type x)
    in
    let path = Filename.concat dir filename in
    if Sys.file_exists path then (
      match ask with
	| Overwrite -> ()
	| Inform f -> f path
	| Stop ->
          Sys.remove file  ;
	  raise (EcurlExeception (File_exists,path) )
    );
    Unix.rename file path
end
