open Pipe

let run_mutexed m f x =
  let module T = struct
    type 'a t =
    | Ok of 'a
    | Err of exn
  end in
  Mutex.lock m;
  let erg = try T.Ok (f x) with | x -> T.Err x in
  Mutex.unlock m;
  match erg with
  | T.Ok x -> x
  | T.Err x -> raise x

type pipe_state =
| Open
| Closed
| Uninit

type pipe_with_status =
  {
    mutable state: pipe_state;
    mutable fd: Unix.file_descr
 }

let pipe a b =
  let tmp1,tmp2 = Unix.pipe () in
  a.state <- Open;
  a.fd <- tmp1;
  b.state <- Open;
  b.fd <- tmp2

let new_pipe () =
  {state = Uninit;
   fd = Unix.stderr}

let close_pipe a =
  match a.state with
  | Closed
  | Uninit -> ()
  | Open ->
    a.state <- Closed ;
    let rec close fd =
      let closed =
        try
          Unix.close fd;
          true
        with
        | Unix.Unix_error(Unix.EINTR,_,_) -> false
      in
      match closed with
      | true -> ()
      | false -> close fd
    in
    close a.fd


type io_out = [
| `Null
| `Stdout
| `Stderr
| `Buffer of Buffer.t
| `Fun of (string -> unit)]

type io_in =
[`String of string
| `Null ]


type t = {
  env: string array;
  stdin: io_in;
  stdout: io_out;
  stderr: io_out;
  prog: string;
  args: string array;
  m: Mutex.t;
  mutable to_kill: bool;
  has_stopped: Condition.t ;
  m_stopped: Mutex.t;
  mutable ret: int ;
  mutable running: bool;
  }


external winkill : int -> unit = "winkill"

let unixkill x = Unix.kill x 15

let kill = match Sys.os_type with
| "Win32" -> winkill
| _ ->  unixkill


let close_pipe_ne a = try close_pipe a with |_ -> ()

let str_buffer_len = 65536


(* counter of running processes, so that main
   program can wait until each has finished *)
let incr_proc_count,
  decr_proc_count,
  wait_exit =
  let is_empty = Condition.create ()
  and m = Mutex.create ()
  and proc_count = ref 0 in
  let incr_proc_count () =
    Mutex.lock m;
    incr proc_count;
    Mutex.unlock m
  and decr_proc_count () =
    Mutex.lock m;
    decr proc_count ;
    if !proc_count <= 0 then
      Condition.broadcast is_empty ;
    Mutex.unlock m;
  and wait_exit () =
    Mutex.lock m;
    while !proc_count > 0 do
      Condition.wait is_empty m
    done;
    Mutex.unlock m
  in
  incr_proc_count,decr_proc_count,wait_exit



(* look each kill_wait_time seconds if the process should be killed *)
let kill_wait_time = 0.5

let run com =
  try
    let args = com.args
    and prog = com.prog
    and stderr = com.stderr
    and stdout = com.stdout
    and stdin = com.stdin
    and env = com.env
    and tmp_str = String.create str_buffer_len
    and p_stdout_read = new_pipe ()
    and p_stdout_write = new_pipe ()
    and p_stderr_read = new_pipe ()
    and p_stderr_write = new_pipe ()
    and p_stdin_read = new_pipe ()
    and p_stdin_write = new_pipe ()
    in
    Util.try_finally ( fun () ->
      pipe p_stdout_read p_stdout_write;
      pipe p_stderr_read p_stderr_write;
      pipe p_stdin_read  p_stdin_write;
      (*let args = match Array.to_list args with
      | [] -> []
      | _::tl -> tl
      in
      let pid = Winproc.create_process ~env
        ~stdin:p_stdin_read.fd
        ~stdout:p_stdout_write.fd
        ~stderr:p_stderr_write.fd
        ~prog ~args
        ()
      in*)
      let pid = Unix.create_process_env prog args env p_stdin_read.fd p_stdout_write.fd p_stderr_write.fd in
      incr_proc_count () ;
      Util.try_finally ( fun () ->
        close_pipe p_stdout_write;
        close_pipe p_stdin_read;
        close_pipe p_stderr_write;

        let f_read r =
          let is_stdout =
            if r = p_stderr_read.fd then
              false
            else (
              assert ( r = p_stdout_read.fd );
              true
            )
          in
          let rec read fd str start len =
            let n = try Unix.read fd str start len with Unix.Unix_error(Unix.EINTR,_,_) -> min_int in
            if n = min_int then read fd str start len else n
          in
          let x = try read r tmp_str 0 str_buffer_len with | _ -> -1 in
          if x <= 0 then (
            if is_stdout then
              close_pipe p_stdout_read
            else
              close_pipe p_stderr_read
          )
          else (
            try
              match if is_stdout then stdout else stderr with
              | `Null   -> ()
              | `Stdout -> String.sub tmp_str 0 x |> print_string
              | `Stderr -> String.sub tmp_str 0 x |> prerr_string
              | `Buffer b -> Buffer.add_substring b tmp_str 0 x
              | `Fun (f: string -> unit) -> f (String.sub tmp_str 0 x)
            with
            | _ -> () (* intentionally catch error in user supplied functions / decr_proc_count *)
          )
        in
        let to_write = match stdin with
        | `String ""
        | `Null -> close_pipe p_stdin_write; ref ""
        | `String str -> ref str
        in
        while p_stdout_read.state = Open || p_stderr_read.state = Open || p_stdin_write.state = Open do
          let wl = if p_stdin_write.state = Open then [p_stdin_write.fd] else [] in
          let rl = if p_stderr_read.state = Open then [p_stderr_read.fd] else [] in
          let rl = if p_stdout_read.state = Open then p_stdout_read.fd :: rl else rl in
          assert ( rl <> [] );
          match try Some(Unix.select rl wl [] kill_wait_time) with | Unix.Unix_error(Unix.EINTR,_,_) -> None with
          | None -> ()
          | Some(r,w,_) ->
            let to_kill = run_mutexed com.m ( fun c -> c.to_kill ) com in
            match to_kill with
            | true ->
              kill pid ;
              close_pipe_ne p_stdin_read;
              close_pipe_ne p_stdin_write;
              close_pipe_ne p_stdout_read;
              close_pipe_ne p_stdout_write;
              close_pipe_ne p_stderr_write;
              close_pipe_ne p_stderr_read;
            | false ->
              List.iter f_read r ;
              match w with
              | [] -> ()
              | [fd] ->
                assert (p_stdin_write.fd = fd);
                let str_len = String.length !to_write in
                assert (str_len > 0 );

                let rec write fd a i l =
                  let n = try Unix.write fd a i l
                    with | Unix.Unix_error(Unix.EINTR,_,_) -> min_int
                  in
                  if n = min_int then write fd a i l else n
                in
                let n_written = write fd !to_write 0 str_len in
                if n_written >= str_len then
                  close_pipe p_stdin_write
                else
                  to_write := String.sub !to_write n_written (str_len - n_written)
              | _ -> assert false
        done;

        close_pipe p_stdout_read;
        close_pipe p_stderr_read;

        let rec waitpid pid =
          let t =
            try
              Some ( Unix.waitpid [] pid )
            with
            | Unix.Unix_error(Unix.EINTR,_,_) -> None
          in
          match t with
          | Some x -> x
          | None -> waitpid pid
        in
        let pid', process_status = waitpid pid in
        assert (pid = pid');
        let ret_code = match process_status with
        | Unix.WEXITED n -> n
        | Unix.WSIGNALED _ -> 2 (* like OCaml's uncaught exceptions *)
        | Unix.WSTOPPED _ ->
        (* only possible if the call was done using WUNTRACED
           or when the child is being traced *)
          3
        in
        com.ret <- ret_code ;
      ) () decr_proc_count ()
    ) () ( fun () ->
      close_pipe_ne p_stdin_read;
      close_pipe_ne p_stdin_write;
      close_pipe_ne p_stdout_read;
      close_pipe_ne p_stdout_write;
      close_pipe_ne p_stderr_write;
      close_pipe_ne p_stderr_read ) ()
  with
  | x ->
    com.ret <- 127;
    raise x

let run com =
  Util.try_finally run com ( fun c ->
    run_mutexed c.m_stopped ( fun c -> c.running <- false; Condition.broadcast c.has_stopped ) c ) com


let kill com =  run_mutexed com.m ( fun c -> c.to_kill <- true ) com
let run com = ignore( Thread.create run com )

let close com =
  Mutex.lock com.m_stopped ;
  while com.running do
    Condition.wait com.has_stopped com.m_stopped ;
  done;
  Mutex.unlock com.m_stopped ;
  com.ret


let new_command ?(env=(Unix.environment ())) ?(stdin=`Null) ?(stdout=`Stdout)
    ?(stderr=`Stderr) (prog:string) (args:string list) =
  {
    env ;
    stdin;
    stdout;
    stderr;
    prog;
    args = Array.of_list ^$ (prog::args)  ;
    m = Mutex.create ();
    to_kill = false ;
    has_stopped = Condition.create ();
    m_stopped = Mutex.create ();
    ret = (-1);
    running = true;
  }
