open Common

let () =
  ignore ( Glib.Timeout.add ~ms:100 ~callback:GtkThread.do_jobs ) ;
  let handle = Sys.Signal_handle (fun signum ->
    Glue.clean_exit () ;
    Run.wait_exit ();
    exit (128 + signum) )
  in
  let help i = try Sys.set_signal i handle with | _ -> () in (* platform dependend, what's supported *)
  let l =  match Const.windows_platform with
  | false -> [ Sys.sigint; Sys.sigterm ; Sys.sigabrt; Sys.sigalrm; Sys.sigfpe;
              Sys.sighup; Sys.sigill; Sys.sigpipe; Sys.sigquit; Sys.sigsegv;  ]
  | true -> [ Sys.sigint ]
  in
  List.iter help l ;
  at_exit Run.wait_exit;
  at_exit Glue.clean_exit


let main_window =
  let title = match Const.windows_platform with
  | true -> "Wodi"
  | false -> "Godi"
  in
  let w = GWindow.window ~title () in
  w#set_icon (Some Const.pixbuf_camel);
  ignore ( w#connect#destroy ~callback:GMain.quit );
  w

let main_vbox = GPack.vbox ~packing:main_window#add ()

let view_source = Listview.s_view
let view_binary = Listview.b_view

let forget_all_changes () =
  let pkgs = !Gl.pkgs
  and to_do = !Gl.to_do in
  for i = 0 to Array.length pkgs - 1 do
    match to_do.(i) with
    | Dont_touch -> ()
    | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall
    | Install | Upgrade | Remove | Reinstall ->
      to_do.(i) <- Dont_touch ;
      Listview.inform i
  done;
  SHashtbl.clear Gl.changed_options;
  Toolbar.apply_button#misc#set_sensitive false ;
  Toolbar.cancel_button#misc#set_sensitive false


let select_newer () =
  let to_do = !Gl.to_do in
  let new_to_do = Array.copy to_do
  and old_to_do = Array.copy to_do
  and changed = ref true in

  let f name =
    try
      let i = SHashtbl.find Gl.names_src name in
      match old_to_do.(i) with
      | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall
      | Install | Upgrade | Remove | Reinstall
        -> ()
      | Dont_touch ->
        new_to_do.(i) <- Upgrade ;
        changed := true
    with
    | Not_found -> Glib.Message.log `WARNING "package %s not found; probably a bug" name ;
  in
  let new_packages = !Gl.glue#get_newer () in
  List.iter f new_packages ;
  match !changed with
  | false ->
    let nothing = "Nothing to do" in
    Info_simple.show_info ~window_title:nothing
      ~title:nothing ~message:"No additional packages have been marked" ()
  | true ->
    match Ginfo.try_apply_to_do new_to_do with
    | false -> ()
    | true ->
      let tree = [
        ("to be rebuild",true,List.sort compare new_packages)
      ]
      and window_title = "Upgrade possible"
      and title = "Upgrade possible"
      and info = "An upgrade is possible. Some packages need to be rebuilt" in
      match Info_tree.ask ~tree ~window_title ~title ~info () with
      | false -> Gl.to_do := old_to_do
      | true ->
        let pkgs = !Gl.pkgs in
        for idx = 0 to Array.length pkgs - 1 do
          if !Gl.to_do.(idx) <> old_to_do.(idx) then
            Listview.inform idx
        done


let update_locale_list ?wl_to_use () =
  Toolbar.update_button#misc#set_sensitive false;
  view_source#deactivate () ;
  view_binary#deactivate ();
  Notebook.unset ();
  Gl.reset ();

  Toolbar.apply_button#misc#set_sensitive false ;
  Toolbar.cancel_button#misc#set_sensitive false ;

  (*Combosearch_source.disconnect ();*)
  Toolbar.combo_source#misc#set_sensitive false;
  Toolbar.combo_bin#misc#set_sensitive false;

  let saved_combo_text_source = Toolbar.combo_source#entry#text
  and saved_combo_text_bin = Toolbar.combo_bin#entry#text in
  Toolbar.combo_source#entry#set_text "";
  Toolbar.combo_bin#entry#set_text "";

  let f_finish_ok () =
    view_source#activate ();
    view_binary#activate ();
    Statusbar.clear ();
    Statusbar.sb#misc#set_sensitive true ;
    Progressbar.deactivate () ;
    Toolbar.update_button#misc#set_sensitive true;
    if !Gl.to_do_cnt > 0 then (
      Toolbar.apply_button#misc#set_sensitive true  ;
      Toolbar.cancel_button#misc#set_sensitive true
    );
    if SHashtbl.length Gl.names_src > 0 then (
      Toolbar.combo_source#misc#set_sensitive true;
      Toolbar.combo_source#entry#set_text saved_combo_text_source ;
      let f cur = match cur.status with
      | Old_version _ -> true
      | Local_version _  | Current_version _  | Unused -> false
      in
      if Util.array_exists f !Gl.pkgs then
        Toolbar.select_newer_button#misc#set_sensitive true;
    );
    if SHashtbl.length Gl.names_bin > 0 then (
      Toolbar.combo_bin#misc#set_sensitive true;
      Toolbar.combo_bin#entry#set_text saved_combo_text_bin ;
    );
    Combosearch.refresh ()

  and f_finish_err () =
    Statusbar.clear ();
    Statusbar.sb#misc#set_sensitive true ;
    Statusbar.flash "Failed loading the package list..." ;
    Toolbar.update_button#misc#set_sensitive true;
    Progressbar.deactivate () ;
  in
  let run () =
    let c = try Some(new Glue.ginfo ~no_source:Const.dont_show_source_packages ()) with _ -> None in
    match c with
    | None -> GtkThread.async f_finish_err ()
    | Some c ->
      let apps = c#get_packages in
      let len = Array.length apps in
      Gl.pkgs :=  apps;
      Gl.to_do := Array.make len Dont_touch;
      Gl.memo_known_dependencies := Array.map ( fun _ -> Hashtbl.create 1 ) apps ;
      for i = 0 to pred len do
        let cur = apps.(i) in
        let name = cur.name in
        match cur.package_type with
        | PSource x ->
          Gl.add_dependencies ~options:x.build_options ~dependencies:cur.dependencies cur;
          SHashtbl.replace Gl.names_src name i
        | PBinary s ->
          Hashtbl.replace Gl.ident_bin s.b_ident i;
          let l = try SHashtbl.find Gl.names_bin name with Not_found -> [] in
          SHashtbl.replace Gl.names_bin name (i::l)
      done;
      Text_index.rebuild ();

      Gl.glue := (c :> glue);

      let wl = match wl_to_use with
      | Some wl -> wl
      | None -> c#get_wishes
      in
      (match c#check_wishes wl with
      | Plan_error _ -> ()
      | (Plan_ok _) as pl_whole ->
        Ginfo.apply_plan pl_whole;
        Ginfo.use_conf_from_wl wl );
      GtkThread.async f_finish_ok ()
  in
  Progressbar.activate () ;
  Statusbar.sb#misc#set_sensitive false ;
  Statusbar.message "Loading the package list...";
  ignore ( Thread.create run () )



let error_message_no_binkkg = "Please make sure that GODI_BINPKG and GODI_BINPKG_SERVER are set in your godi.conf"
let get_local_and_remote suffix =
  let godi_binpkg_server = Const.godi_binpkg_server
  and godi_binpkg = Const.godi_binpkg  in
  match godi_binpkg,godi_binpkg_server with
  | Some "", _
  | _ , Some ""
  | None,_
  |_, None -> None
  | Some(local_address),Some(server_address) ->
    let url = match server_address.[String.length server_address - 1 ] with
    | '/'  -> server_address ^ suffix
    | _    -> server_address ^ "/" ^ suffix
    and file = Filename.concat local_address suffix in
    Some(file,url)


let manifest_file = "PKGMANIFEST"
let update_online () =
  if Const.dont_show_source_packages = true && Const.windows_platform = true then
    match get_local_and_remote manifest_file with
    | None ->
      Info_simple.show_error ~window_title:"Update your godi.conf"
        ~title:"Variables not set" ~message:error_message_no_binkkg ()
    | Some(file,url) ->
      match Download_dialog.dialog ~window_title:"Updating list of binary packages" [(url,file)] with
      | true ->
        let wl_to_use = Ginfo.get_cur_wl () in
        update_locale_list ~wl_to_use ()
      | false ->
        let message = "I could not download a new list of available binary packages. Please try again later." in
        Info_simple.show_error ~window_title:"Update failed" ~title:"Update failed" ~message ()
  else
    let open Console in
    let l = [{
    (*    chdir = None; *)
      command = Const. godi_console ;
      args =  [| "update" |] ;
      killable = true;
      description = "Updating the package list" ;
      env = None;
    }]
    and window_title = "Update the list of avaiable packages" in
    match run ~icon:Const.pixbuf_camel ~auto_close:true ~window_title l with
    | Error _
    | Aborted _ -> ()
    | Ok ->
      let wl_to_use = Ginfo.get_cur_wl () in
      update_locale_list ~wl_to_use ()


let dummy_wish = Wish_option(Const.invalid_string,Const.invalid_string,Const.invalid_string)

exception Download_failed of string
let apply () =
  match !Gl.glue#get_apply_order ^$ Ginfo.get_cur_wl () with
  | Plan_error(message) ->
    let d = GWindow.message_dialog ~message ~message_type:`ERROR ~buttons:GWindow.Buttons.close ~show:true () in
    (match d#run () with
    | _ -> d#destroy () )
  | Plan_ok wl ->
    let make_help ?(killable=true) ?(wish=dummy_wish) ~app ~arg ~desc () =
      let x = {
        Console.command = Const.godi_console ;
        Console.args = [| "make" ; "-C" ; Ptype.get_app_dir app ; "-f" ; "Makefile" ; arg |] ;
        Console.killable = killable;
        Console.description = desc ^ " " ^ app;
        Console.env = Const.sub_prog_env;
      }
      in
      wish,x
    in
    let f_fetch_binaries accu = function
    | Plan_remove _  | Plan_remove_dep _  | Plan_build _  | Plan_build_dep _ -> accu
    | Plan_add x | Plan_add_dep x ->
      match !Gl.glue#get_bin_fln x with
      | None -> raise (Download_failed ( "I don't know how to download the following package: " ^
                                          (Ptype.name_of_bi x ) ));
      | Some fln ->
        match get_local_and_remote fln with
        | None ->
          raise (Download_failed error_message_no_binkkg);
        | Some(local_address,server_address) ->
          if Sys.file_exists local_address then
            accu
          else
            (server_address,local_address)::accu
    and f_fetch_src accu = function
    | Plan_build s
    | Plan_build_dep s ->
      (make_help ~app:s ~arg:"fetch" ~desc:"Fetching " () )::accu
    | Plan_remove _  | Plan_remove_dep _  | Plan_add _ | Plan_add_dep _ -> accu
    and f_apply accu pl = match pl with
    | Plan_build s
    | Plan_build_dep s ->
      let wish = match pl with
      | Plan_build _ -> Wish_build s
      |  _ -> dummy_wish
      in
      (make_help ~app:s ~arg:"clean" ~desc:"Cleaning" () )
      ::(make_help ~wish ~killable:false ~app:s ~arg:"install" ~desc:"installing" () )
      ::(make_help ~app:s ~arg:"build" ~desc:"Building" () )
      ::(make_help ~app:s ~arg:"configure" ~desc:"Configuring" () )
      ::(make_help ~app:s ~arg:"patch" ~desc:"Patching" () )
      ::(make_help ~app:s ~arg:"extract" ~desc:"Extracting" () )
      ::(make_help ~app:s ~arg:"clean" ~desc:"Cleaning for" () )
      ::accu
    | Plan_remove s
    | Plan_remove_dep s ->
      let name = Ptype.name_of_bi s in
      let wish = match pl with
      | Plan_remove _ -> Wish_remove name
      | _ -> dummy_wish
      in
      (wish,{
        Console.command =  Const.godi_console ;
        Console.args =  [| "delete" ; "-v" ; name |] ;
        Console.killable = false;
        Console.description = "Removing " ^ name;
        Console.env = Const.sub_prog_env;
      })::accu
    | Plan_add s
    | Plan_add_dep s ->
      let wish = match pl with
      | Plan_add _ -> Wish_add s
      | _ -> dummy_wish
      in
      let name = Ptype.name_of_bi s in
      let arg3 = match !Gl.glue#get_bin_fln s with
      | None -> name
      | Some fln ->
        match Glue.godi_binpkg () with
        | None -> name
        | Some x -> Filename.concat x fln
      in
      (wish,{
        Console.command =  Const.godi_console ;
        Console.args =  [| "add" ; "-v" ; arg3 |];
        Console.killable = false;
        Console.description = "Adding " ^ name ;
        Console.env = Const.sub_prog_env;
      })::accu
    in
    try
      begin
      if Const.windows_platform = true then
        (* windows specific addition *)
        match List.fold_left f_fetch_binaries [] wl |> List.rev with
        | [] -> ()
        | l ->
          match Download_dialog.dialog ~window_title:"Fetching binary packages" l with
          | true -> ()
          | false -> raise (Download_failed "I could not download the necessary binary packages from the remote server")
      end;
      let l = List.fold_left f_fetch_src [] wl in
      let l = List.rev ^$ List.fold_left f_apply l wl  in
      let window_title = "Apply the changes"
      and success_message = Const.apply_changes_success_message in
      let wl_to_use =
        match List.map snd l |>
            Console.run ~icon:Const.pixbuf_camel ~auto_close:false ~success_message ~window_title
        with
        | Console.Ok -> None
        | Console.Error tl
        | Console.Aborted tl ->
          let len = List.length l - List.length tl in
          let (_,still_to_do) = Util.list_partition_by_number l len in
          match List.filter ( (!=) dummy_wish ) ^$ List.map fst still_to_do with
          | [] -> None
          | l -> Some l
      in
      update_locale_list ?wl_to_use ()
    with
    | Download_failed message ->
      Info_simple.show_error ~window_title:"Download failed" ~title:"Download failed!" ~message ()



let nb_src_bin = GPack.notebook ~tab_pos:`TOP  ()
let switch_page = function
| 0 -> Toolbar.combo_source#misc#show () ; Toolbar.combo_bin#misc#hide ()
| 1 -> Toolbar.combo_source#misc#hide () ; Toolbar.combo_bin#misc#show ()
| _ -> assert false


let init_main () =
  let vpaned = GPack.paned `VERTICAL ()
  and sw_package_src = GBin.scrolled_window ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ()
  and sw_package_bin = GBin.scrolled_window ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ()  in

  (match Const.dont_show_source_packages with
  | true ->
    Toolbar.combo_source#misc#hide () ; Toolbar.combo_bin#misc#show ();
    vpaned#add1 sw_package_bin#coerce
  | false ->
    vpaned#add1 nb_src_bin#coerce ;
    nb_src_bin#misc#set_can_focus false ;
    ignore ( nb_src_bin#connect#switch_page ~callback:switch_page );
    let mlabel text =
      let t = GMisc.label ~text () in
    (* t#set_angle 180. ; *)
      t#misc#set_can_focus false ;
      t#coerce
    in
    let _ = nb_src_bin#append_page ~tab_label:(mlabel "Source") sw_package_src#coerce in
    let _ = nb_src_bin#append_page ~tab_label:(mlabel "Binary") sw_package_bin#coerce in
    ()
  );

  vpaned#add2  Notebook.sw#coerce ;

  main_vbox#pack ~expand:false ~fill:false Toolbar.tb#coerce;
  main_vbox#pack ~expand:true ~fill:true vpaned#coerce ;

  sw_package_src#add view_source#coerce;
  sw_package_bin#add view_binary#coerce;

  let width =
    let screen_width = float_of_int (Gdk.Screen.width ()) in
    let tmp = screen_width *. 0.75 in
    int_of_float ( if tmp > 800. then tmp else screen_width -. 20. )
  and height =
    let screen_height = float_of_int (Gdk.Screen.height ()) in
    let tmp = screen_height *. 0.75 in
    int_of_float ( if tmp > 640. then tmp /. 2.  else (screen_height -. 20.) /.  2. )
  in
  sw_package_src#misc#set_size_request ~height ~width () ;
  sw_package_bin#misc#set_size_request ~height ~width () ;
  Notebook.sw#misc#set_size_request ~height ~width ();


  main_vbox#pack (*~expand:false*) Statusbar.sb#coerce ;

  let ali = GBin.alignment ~xalign:1. ~yalign:0. ~packing:Statusbar.sb#pack () in
  ali#add Progressbar.pb#coerce ;
  main_window#show () ;

  let _ = Toolbar.cancel_button#connect#clicked ~callback:forget_all_changes in
  let _ = Toolbar.update_button#connect#clicked ~callback:update_online in
  let _ = Toolbar.apply_button#connect#clicked ~callback:apply in
  let _ = Toolbar.select_newer_button#connect#clicked ~callback:select_newer in

  update_locale_list ();

  Toolbar.about#misc#set_sensitive true


let () =
  init_main ();
  GtkThread.main ()
