open Common

let sw = GBin.scrolled_window ~border_width:0 ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ()

let frame =
  let f = GBin.frame (*~border_width:2 ~shadow_type:`IN*) () in
  sw#add_with_viewport f#coerce ;
  f

let nb = GPack.notebook ~packing:frame#add ()


let pango_mono = Pango.Font.from_string "monospace"

(* workaround for french (and german) names *)
let to_utf8 str =
  try
    Glib.Convert.convert_with_fallback ~to_codeset:"utf8" ~from_codeset:"iso8859-1" str
  with
  | _ ->
    try
      Glib.Convert.convert_with_fallback ~to_codeset:"utf8" ~from_codeset:"iso8859-15" str
    with
    | _ -> str

type t = {
  key: GMisc.label;
  value: GMisc.label
}

let label_desc ?(line_wrap=true) text =
  let l1 = GMisc.label ~yalign:0. ~xalign:0. ~selectable:true ~line_wrap:true ~text ()
  and l2 = GMisc.label ~yalign:0. ~xalign:0. ~selectable:true ~line_wrap ~text:"" ()  in
  l1#misc#set_can_focus false ;
  l2#misc#set_can_focus false ;
  { key = l1 ; value =l2 }


let string_installed_version cur =
  let normal = match cur.status with
  | Current_version x
  | Local_version x
  | Old_version x -> Some x.i_pretty_version
  | Unused -> None
  in
  match normal with
  | None -> ""
  | Some v ->
    match cur.package_type with
    | PSource _ -> v
    | PBinary x -> v ^ "@" ^ x.build_version_fingerprint

let string_version cur =
  let v = cur.pretty_version in
  match cur.package_type with
  | PSource _ -> v
  | PBinary x -> v ^ "@" ^ x.build_version_fingerprint


(*  in
  match cur.package_type with
  | PSource -> first
  | _ when cur.status = Unused -> first
  | PBinary ->
    let second = match cur.status with
    | Unused -> assert false
    | Current_version x
    | Local_version x
    | Old_version x -> x.i_pretty_version
    in
    first ^ "@" ^ second
*)


(*let computing_message = "still computing ..."*)


class build_page = object
  val table = GPack.table ~col_spacings:5 ~row_spacings:3 ~border_width:5 ~rows:4 ~columns:2 ()

  val version = label_desc "Build version:"
  val fp = label_desc "Fingerprint:"
  val info = label_desc "Info:"
  val imessage = label_desc "Install message:"


  method widget = table#coerce

  method set i =
    let cur = !Gl.pkgs.(i) in
    let bi = Ptype.get_binary_info cur in

    begin match bi with
    | None ->
      version.value#set_text "-" ;
      fp.value#set_text "-";
      info.value#set_text "-";
    | Some i ->
      let text = match i.build_version with
      | None -> "-"
      | Some x -> to_utf8 ^$ Ptype.string_of_version x
      in
      version.value#set_text text ;
      fp.value#set_text ^$ to_utf8 i.build_version_fingerprint;
      let text = match i.build_info with
      | None -> "-"
      | Some x -> to_utf8 x
      in
      info.value#set_text text
    end;

    begin match cur.install_message with
    | "" ->
      imessage.key#misc#hide () ;
      imessage.value#misc#hide () ;
      imessage.value#set_text "-"
    | x ->
      imessage.key#misc#show () ;
      imessage.value#misc#show () ;
      imessage.value#set_text ^$ to_utf8 x
    end


  initializer
  let i = ref 0 in
  let help cur =
    table#attach ~left:0 ~top:!i (cur.key#coerce);
    table#attach ~left:1 ~top:!i (cur.value#coerce);
    incr i
  in
  help version ;
  help fp;
  help info;
  help imessage;

  version.value#misc#modify_font pango_mono ;
  info.value#misc#modify_font pango_mono


end

let build_page = new build_page

class desc_page = object
  val table = GPack.table ~col_spacings:5 ~row_spacings:3 ~border_width:5 ~rows:7 ~columns:2 ()

  val cur_version = label_desc "Installed version:"
  val avaiable_version = label_desc "Avaiable version:"
  val comment = label_desc "Comment:"
  val description = label_desc ~line_wrap:false "Description:"
  val dependencies = label_desc "Dependencies:"
  val filename = label_desc "Filename:"
  val maintainer = label_desc "Maintainer:"
  val homepage = label_desc "Homepage:"

  val mutable cur_name = ""
  val mutable cur_index = (-1)

  method widget = table#coerce

  method update_dependencies i name =
    if cur_index <> i || name <> cur_name then
      false
    else
      let o_dep_str = Ptype.dependencies_to_string !Gl.pkgs.(i).dependencies in
      (*let str = match o_dep_str with
      | Some x -> x
      | None -> computing_message in (* TODO: remove me *) *)
      dependencies.value#set_text o_dep_str;
      (* o_dep_str = None *)
      false


  method set i =
    let cur = !Gl.pkgs.(i) in
    let cdependencies = Ptype.dependencies_to_string cur.dependencies
      (*
      match dependencies_to_string cur "\n" with (* todo: fall, wenn noch nicht berechnet! *)
    | Some x -> x
    | None -> computing_message *)
    and chomepage = Ginfo.get_homepage i
    and cmaintainer = match cur.godi_maintainer with
    | "" -> None
    | x -> Some x in

    cur_name <- cur.name;
    cur_index <- i;

    cur_version.value#set_text ^$ string_installed_version cur ;
    avaiable_version.value#set_text ^$ string_version cur;
    comment.value#set_text ^$ to_utf8 cur.comment;
    description.value#set_text ^$ to_utf8 cur.description ;

    (match cur.package_type with
     | PSource _ -> filename.key#misc#hide (); filename.value#misc#hide ()  ;
     | PBinary x ->
        match !Gl.glue#get_bin_fln x.b_ident with
        | None -> filename.key#misc#hide (); filename.value#misc#hide ()  ;
        | Some x ->
           filename.value#set_text x ;
           filename.key#misc#show () ; filename.value#misc#show ()  ;
    ) ;




    dependencies.value#set_text cdependencies;
    (match cmaintainer with
    | None -> maintainer.value#set_text "-"
    | Some x -> maintainer.value#set_text ^$ to_utf8 x);
    (match chomepage with
    | None -> homepage.value#set_text "-"
    | Some x -> homepage.value#set_text ^$ to_utf8 x)

  initializer
  let i = ref 0 in
  let help cur =
    table#attach ~left:0 ~top:!i (cur.key#coerce);
    table#attach ~left:1 ~top:!i (cur.value#coerce);
    incr i
  in
  help cur_version ;
  help filename ;
  help avaiable_version;
  help comment;
  help description;
  help dependencies ;
  help maintainer;
  help homepage;
  description.value#misc#modify_font pango_mono
end

let desc_page = new desc_page



let make_conf_page index () =
  let app = !Gl.pkgs.(index) in
  let app_name = app.name in

  let app_cur_build_options = Gl.get_cur_build_options app in
  let tmp_options =
    let s = SHashtbl.create ^$ List.length app_cur_build_options in
    List.iter ( fun (k,v) -> SHashtbl.replace s k v ) app_cur_build_options;
    s
  in

  let config_callback () =
    let dir = Ptype.get_app_dir app_name in
    let makefile = Filename.concat dir "Makefile" in

    let l = [{
      (*Console.chdir = Some(get_app_dir app_name);*)
      Console.command =  Const.godi_console ;
      Console.args =  Array.of_list ^$ List.rev ^$
        SHashtbl.fold ( fun k v a -> ( k ^ "=" ^ v )::a ) tmp_options
        ( "configure"::"clean"::makefile::"-f"::dir::"-C"::"make"::[] );
      Console.killable = true;
      Console.description = Printf.sprintf "Configuring %s" app_name ;
      Console.env = Const.sub_prog_env;
    }]
    and window_title = "Test configuration of " ^ app_name in
    let _ = Console.run ~icon:Const.pixbuf_camel
      ~success_message:Const.config_success_message ~window_title l
    in
    ()
  in

  let test_button = GButton.button ~label: "Test configuration" () in
  ignore ( test_button#connect#clicked ~callback:config_callback );

  let options = match app_cur_build_options with
  | [] -> test_button#coerce
  | _ ->
    let previous_options = app_cur_build_options
    and htl_entries = SHashtbl.create ^$ List.length app_cur_build_options
    and frame = GBin.frame ~border_width:10 ~shadow_type:`IN () in
    let vbox = GPack.vbox ~border_width:4 ~packing:frame#add () in

    let _i:int =
      let bo = Gl.get_cur_build_options app in
      let t = GPack.table ~row_spacings:3 ~col_spacings:5 ~border_width:4 ~rows:(List.length bo)
        ~columns:2  ~packing:vbox#add  ()
      in
      List.fold_left ( fun i (key,text) ->
        let l = GMisc.label ~yalign:0.5 ~xalign:0. ~selectable:true ~text:(key ^ ":") ()
        and e = GEdit.entry ~editable:true ~text ~width:130 ~show:true () in
        e#misc#set_can_focus true ;
        l#misc#set_can_focus false ;
        t#attach ~left:0 ~top:i l#coerce ;
        t#attach ~left:1 ~top:i e#coerce ;
        SHashtbl.replace htl_entries key e ;
        let callback () = SHashtbl.replace tmp_options key e#text in
        let _ = e#connect#changed ~callback in
        succ i
      ) 0 bo
    in

    let apply_button = GButton.button ~label: "Apply" () in

    let shtl_changed =  Gl.get_changed_options app  in
    let shtl_changed_orig = SHashtbl.copy shtl_changed in
    let show er_mgs =
      (match er_mgs with
      | None ->
        let f k v =
          SHashtbl.replace shtl_changed  k v ;
          SHashtbl.replace shtl_changed_orig k v;
        in
        SHashtbl.iter f tmp_options ;
        ignore (desc_page#update_dependencies index app.name )
      | Some x ->
        let message = "I can't apply the changes: \n\n" ^ x in
        let d = GWindow.message_dialog ~message ~message_type:`ERROR ~buttons:GWindow.Buttons.close ~show:true () in
        (match d#run () with
        | _ -> d#destroy (); );
        let f (k,v) =
          let e = SHashtbl.find htl_entries k in
          e#set_text v ;
          if SHashtbl.mem shtl_changed_orig k then
            SHashtbl.replace shtl_changed k v
        in
        List.iter f previous_options
      );
      nb#misc#set_sensitive true ;
      Toolbar.tb#misc#set_sensitive true;
      Progressbar.deactivate () ;
      Statusbar.clear ();
      if er_mgs = None then
        Statusbar.flash "Configuration applied"
    in
    let compute () =
      let f k v =
          (* SHashtbl.replace app_cur_build_options k v ; *)
        SHashtbl.replace shtl_changed k v
      in
      SHashtbl.iter f tmp_options ;
      let er_mgs = match  !Gl.to_do.(index) with
      | Dont_touch -> Ginfo.compute_depends app ; None
      | Depending_install | Depending_upgrade | Depending_remove | Depending_reinstall
      | Install | Upgrade | Remove | Reinstall
        ->
        let wl = Ginfo.get_cur_wl () in
        match Ginfo.check_plan ~force_rescan:true ~wl () with
        | Plan_error(message) -> Some(message)
        | (Plan_ok pl) as pl_whole  ->
          let to_do = !Gl.to_do in
          match Ginfo.check_plan_local ~to_do ~wl ~pl with
          | [] ->
            Ginfo.apply_plan pl_whole;
            Ginfo.compute_depends app ;
            None
          |  l ->
            let message = String.concat "\n" ^$ List.map snd l in
            Some(message)
      in
      (match er_mgs with
      | Some _ -> ()
      | None ->
        let options = SHashtbl.fold ( fun k v a -> (k,v)::a ) tmp_options [] in
        !Gl.glue#remember_config app.name  ~options
      );
      GtkThread.async show er_mgs
    in

    let _ = apply_button#connect#clicked ~callback:( fun () ->
      nb#misc#set_sensitive false ;
      Toolbar.tb#misc#set_sensitive false;
      Progressbar.activate () ;
      Statusbar.message "Apply configuration...";
      ignore ( Thread.create compute () )
    )
    in

    let hbox = GPack.hbox ~border_width:4 ~packing:vbox#add () in

    let ali = GBin.alignment ~xalign:0. ~xscale:0. ~packing:hbox#add () in
    ali#add apply_button#coerce ;

    let ali = GBin.alignment ~xalign:1. ~xscale:0. ~packing:hbox#add () in
    ali#add test_button#coerce;

    frame#coerce

  and final_desc =
    let text = match app.package_type with
    | PBinary _ -> assert false
    | PSource x -> x.build_message
    in
    let l = GMisc.label ~yalign:0. ~xalign:0. ~selectable:true ~line_wrap:true ~text () in
    l#misc#set_can_focus false ;
    l#misc#modify_font pango_mono;
    l#coerce

  and table = GPack.table ~row_spacings:9 ~border_width:5 ~rows:2 ~columns:1 () in
  table#attach ~left:0 ~top:0 options;
  table#attach ~left:0 ~top:1 final_desc ;
  table#coerce


let set i =
  let () =
    let further = ref true in
    while !further do
      match nb#current_page with
      | -1 -> further := false;
      | i  -> nb#remove_page i
    done
  in
  if i >= 0 then (
    let cur = !Gl.pkgs.(i) in

    sw#vadjustment#set_value 0.0;
    (*sw#set_vadjustment va; *)

    frame#misc#set_property "label" (`STRING(Some cur.name )) ;
    desc_page#set i;

    ignore ( nb#append_page ~tab_label:(GMisc.label ~text:"Info" () :> GObj.widget ) desc_page#widget );

    (match cur.package_type with
    | PSource _ ->
      let conf_page = make_conf_page i () in
      ignore ( nb#append_page ~tab_label:(GMisc.label ~text:"Configuration" () :> GObj.widget ) conf_page ) ;
    | PBinary _ -> ());

    (match Ptype.get_binary_info cur with
    | None -> ()
    | Some _ ->
      build_page#set i;
      ignore ( nb#append_page ~tab_label:(GMisc.label ~text:"Build" () :> GObj.widget ) build_page#widget ) ;
    );
  )
  else
    frame#misc#set_property "label" (`STRING(None ))

let unset () = set (-1)
(*let notebook = nb*)


(*let nb = frame#coerce*)
